import json
import warnings
import sqlite3
import traceback
import copy
from functools import partial
import os
import yaml

import numpy as np
import pandas as pd
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QGroupBox, QLineEdit, QLabel, QPushButton
from PyQt5.QtWidgets import QListWidgetItem, QMessageBox, QWidget
from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout
from PyQt5.QtWidgets import QTableWidgetItem, QPlainTextEdit, QSizePolicy
from coolname import generate_slug
from pydantic import ValidationError
from xopt import VOCS
from xopt.generators import get_generator_defaults, all_generator_names
from xopt.utils import get_local_region

from badger.gui.default.components.generator_cbox import BadgerAlgoBox
from badger.gui.default.components.data_table import (
    get_table_content_as_dict,
    set_init_data_table,
    update_init_data_table,
)
from badger.gui.default.components.env_cbox import BadgerEnvBox
from badger.gui.default.components.filter_cbox import BadgerFilterBox
from badger.gui.default.components.state_item import state_item
from badger.gui.default.windows.docs_window import BadgerDocsWindow
from badger.gui.default.windows.env_docs_window import BadgerEnvDocsWindow
from badger.gui.default.windows.edit_script_dialog import BadgerEditScriptDialog
from badger.gui.default.windows.lim_vrange_dialog import BadgerLimitVariableRangeDialog
from badger.gui.default.windows.review_dialog import BadgerReviewDialog
from badger.gui.default.windows.add_random_dialog import BadgerAddRandomDialog
from badger.gui.default.windows.message_dialog import BadgerScrollableMessageBox
from badger.gui.default.windows.expandable_message_box import ExpandableMessageBox
from badger.gui.default.utils import filter_generator_config
from badger.db import save_routine, update_routine, get_runs_by_routine
from badger.environment import instantiate_env
from badger.errors import BadgerRoutineError
from badger.factory import list_generators, list_env, get_env
from badger.routine import Routine
from badger.settings import init_settings
from badger.utils import (
    get_yaml_string,
    load_config,
    strtobool,
    get_badger_version,
    get_xopt_version,
)


class BadgerRoutinePage(QWidget):
    sig_updated = pyqtSignal(str, str)  # routine name, routine description

    def __init__(self):
        super().__init__()

        self.generators = list_generators()
        self.envs = list_env()
        self.env = None
        self.routine = None
        self.script = ""
        self.window_docs = BadgerDocsWindow(self, "")
        self.window_env_docs = BadgerEnvDocsWindow(self, "")
        self.vars_env = None  # needed for passing env vars to the var table

        # Limit variable ranges
        self.limit_option = {
            "limit_option_idx": 0,
            "ratio_curr": 0.1,
            "ratio_full": 0.1,
        }

        # Add radom points config
        self.add_rand_config = {
            "method": 0,
            "n_points": 3,
            "fraction": 0.1,
        }
        self.rc_dialog = None

        # Record the initial table actions
        self.init_table_actions = []
        # Record the ratio var ranges
        self.ratio_var_ranges = {}

        self.init_ui()
        self.config_logic()

        # Trigger the re-rendering of the environment box
        self.env_box.relative_to_curr.setChecked(True)

    def init_ui(self):
        config_singleton = init_settings()

        # Set up the layout
        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(11, 11, 19, 11)

        # Meta group
        group_meta = QGroupBox("Metadata")
        group_meta.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        vbox_meta = QVBoxLayout(group_meta)

        # Name
        name = QWidget()
        hbox_name = QHBoxLayout(name)
        hbox_name.setContentsMargins(0, 0, 0, 0)
        label = QLabel("Name")
        label.setFixedWidth(64)
        self.edit_save = edit_save = QLineEdit()
        edit_save.setPlaceholderText(generate_slug(2))
        hbox_name.addWidget(label)
        hbox_name.addWidget(edit_save, 1)
        vbox_meta.addWidget(name, alignment=Qt.AlignTop)

        # Description
        descr = QWidget()
        hbox_descr = QHBoxLayout(descr)
        hbox_descr.setContentsMargins(0, 0, 0, 0)
        lbl_descr_col = QWidget()
        vbox_lbl_descr = QVBoxLayout(lbl_descr_col)
        vbox_lbl_descr.setContentsMargins(0, 0, 0, 0)
        lbl_descr = QLabel("Descr")
        lbl_descr.setFixedWidth(64)
        vbox_lbl_descr.addWidget(lbl_descr)
        vbox_lbl_descr.addStretch(1)
        hbox_descr.addWidget(lbl_descr_col)

        edit_descr_col = QWidget()
        vbox_descr_edit = QVBoxLayout(edit_descr_col)
        vbox_descr_edit.setContentsMargins(0, 0, 0, 0)
        self.edit_descr = edit_descr = QPlainTextEdit()
        edit_descr.setMaximumHeight(80)
        vbox_descr_edit.addWidget(edit_descr)
        descr_bar = QWidget()
        hbox_descr_bar = QHBoxLayout(descr_bar)
        hbox_descr_bar.setContentsMargins(0, 0, 0, 0)
        self.btn_descr_update = btn_update = QPushButton("Update Description")
        btn_update.setDisabled(True)
        btn_update.setFixedSize(128, 24)
        hbox_descr_bar.addStretch(1)
        hbox_descr_bar.addWidget(btn_update)
        vbox_descr_edit.addWidget(descr_bar)
        hbox_descr.addWidget(edit_descr_col)
        vbox_meta.addWidget(descr)

        # Tags
        self.cbox_tags = cbox_tags = BadgerFilterBox(title=" Tags")
        if not strtobool(config_singleton.read_value("BADGER_ENABLE_ADVANCED")):
            cbox_tags.hide()
        vbox_meta.addWidget(cbox_tags, alignment=Qt.AlignTop)
        # vbox_meta.addStretch()

        vbox.addWidget(group_meta)

        # Algo box
        self.generator_box = BadgerAlgoBox(None, self.generators)
        self.generator_box.expand()  # expand the box initially
        vbox.addWidget(self.generator_box)

        # Env box
        BADGER_PLUGIN_ROOT = config_singleton.read_value("BADGER_PLUGIN_ROOT")
        env_dict_dir = os.path.join(
            BADGER_PLUGIN_ROOT, "environments", "env_colors.yaml"
        )
        try:
            with open(env_dict_dir, "r") as stream:
                env_dict = yaml.safe_load(stream)
        except (FileNotFoundError, yaml.YAMLError):
            env_dict = {}
        self.env_box = BadgerEnvBox(env_dict, None, self.envs)
        self.env_box.expand()  # expand the box initially
        vbox.addWidget(self.env_box)

        vbox.addStretch()

    def config_logic(self):
        self.btn_descr_update.clicked.connect(self.update_description)
        self.generator_box.cb.currentIndexChanged.connect(self.select_generator)
        self.generator_box.btn_docs.clicked.connect(self.open_generator_docs)
        self.generator_box.check_use_script.stateChanged.connect(self.toggle_use_script)
        self.generator_box.btn_edit_script.clicked.connect(self.edit_script)
        self.env_box.cb.currentIndexChanged.connect(self.select_env)
        self.env_box.btn_env_play.clicked.connect(self.open_playground)
        self.env_box.btn_docs.clicked.connect(self.open_environment_docs)
        self.env_box.btn_add_var.clicked.connect(self.add_var)
        self.env_box.btn_lim_vrange.clicked.connect(self.limit_variable_ranges)
        self.env_box.btn_add_con.clicked.connect(self.add_constraint)
        self.env_box.btn_add_sta.clicked.connect(self.add_state)
        self.env_box.btn_add_curr.clicked.connect(
            partial(self.fill_curr_in_init_table, record=True)
        )
        self.env_box.btn_add_rand.clicked.connect(self.show_add_rand_dialog)
        self.env_box.btn_clear.clicked.connect(
            partial(self.clear_init_table, reset_actions=True)
        )
        self.env_box.btn_add_row.clicked.connect(self.add_row_to_init_table)
        self.env_box.relative_to_curr.stateChanged.connect(self.toggle_relative_to_curr)
        self.env_box.var_table.sig_sel_changed.connect(self.update_init_table)
        self.env_box.var_table.sig_pv_added.connect(self.handle_pv_added)

    def refresh_ui(self, routine: Routine = None, silent: bool = False):
        self.routine = routine  # save routine for future reference

        self.generators = list_generators()
        self.envs = list_env()
        # Clean up the constraints/observables list
        self.env_box.con_table.clear_constraints()
        self.env_box.list_obs.clear()

        if routine is None:
            # Reset the generator and env configs
            self.generator_box.cb.setCurrentIndex(-1)
            self.env_box.cb.setCurrentIndex(-1)
            init_table = self.env_box.init_table
            init_table.clear()
            init_table.horizontalHeader().setVisible(False)
            init_table.setRowCount(10)
            init_table.setColumnCount(0)

            # Reset the routine configs check box status
            self.env_box.check_only_var.setChecked(False)
            self.env_box.check_only_obj.setChecked(False)
            self.env_box.relative_to_curr.setChecked(True)
            self.try_populate_init_table()

            # Reset the save settings
            name = generate_slug(2)
            self.edit_save.setText("")
            self.edit_save.setPlaceholderText(name)
            self.edit_descr.setPlainText("")
            self.btn_descr_update.setDisabled(True)

            return

        # Enable description edition
        self.btn_descr_update.setDisabled(False)
        # Fill in the generator and env configs
        name_generator = routine.generator.name
        try:
            idx_generator = self.generators.index(name_generator)
        except ValueError as e:
            if not silent:  # show the error message if not in silent mode
                details = traceback.format_exc()
                dialog = BadgerScrollableMessageBox(
                    title="Error!", text=str(e), parent=self
                )
                dialog.setIcon(QMessageBox.Critical)
                dialog.setDetailedText(details)
                dialog.exec_()

            idx_generator = -1

        self.generator_box.cb.setCurrentIndex(idx_generator)
        # self.generator_box.edit.setPlainText(routine.generator.yaml())
        filtered_config = filter_generator_config(
            name_generator, routine.generator.model_dump()
        )
        self.generator_box.edit.setPlainText(get_yaml_string(filtered_config))
        self.script = routine.script

        name_env = routine.environment.name
        idx_env = self.envs.index(name_env)
        self.env_box.cb.setCurrentIndex(idx_env)
        env_params = routine.environment.model_dump()
        del env_params["interface"]
        self.env_box.edit.setPlainText(get_yaml_string(env_params))

        # Config the vocs panel
        variables = routine.vocs.variable_names
        self.env_box.check_only_var.setChecked(True)

        self.env_box.edit_var.clear()

        # Add additional variables to table as well
        # Combine the variables from the env with the additional variables
        all_variables = {}  # note this stores the hard bounds of the variables
        for i in self.vars_env:
            all_variables.update(i)
        if routine.additional_variables:  # there are additional variables
            env = self.create_env()
            # Have to check each variable since some could fail
            for v in routine.additional_variables:
                try:
                    b = env.get_bound(v)
                except Exception:
                    b = [-1000, 1000]  # default wide range
                all_variables.update({v: b})
        # Format for update_variables method
        all_variables = dict(sorted(all_variables.items()))
        all_variables = [{key: value} for key, value in all_variables.items()]

        self.env_box.var_table.update_variables(all_variables)
        self.env_box.var_table.set_selected(variables)
        self.env_box.var_table.addtl_vars = routine.additional_variables

        flag_relative = routine.relative_to_current
        if flag_relative:  # load the relative to current settings
            self.ratio_var_ranges = routine.vrange_limit_options
            self.init_table_actions = routine.initial_point_actions
        self.env_box.relative_to_curr.blockSignals(True)
        self.env_box.relative_to_curr.setChecked(flag_relative)
        self.env_box.relative_to_curr.blockSignals(False)
        self.toggle_relative_to_curr(flag_relative, refresh=False)

        # Always use ranges stored in routine
        self.env_box.var_table.set_bounds(routine.vocs.variables, signal=False)

        # Fill in initial points stored in routine if available
        try:
            init_points = routine.initial_points
            set_init_data_table(self.env_box.init_table, init_points)
        except KeyError:
            set_init_data_table(self.env_box.init_table, None)

        objectives = routine.vocs.objective_names
        self.env_box.check_only_obj.setChecked(True)
        self.env_box.edit_obj.clear()
        self.env_box.obj_table.set_selected(objectives)
        self.env_box.obj_table.set_rules(routine.vocs.objectives)

        constraints = routine.vocs.constraints
        if len(constraints):
            for name, val in constraints.items():
                relation, thres = val
                critical = name in routine.critical_constraint_names
                relation = ["GREATER_THAN", "LESS_THAN", "EQUAL_TO"].index(relation)
                self.add_constraint(name, relation, thres, critical)

        observables = routine.vocs.observable_names
        if len(observables):
            for name_sta in observables:
                self.add_state(name_sta)

        # Config the metadata
        self.edit_save.setPlaceholderText(generate_slug(2))
        self.edit_save.setText(routine.name)
        self.edit_descr.setPlainText(routine.description)

        self.generator_box.check_use_script.setChecked(not not self.script)

    def select_generator(self, i):
        # Reset the script
        self.script = ""
        self.generator_box.check_use_script.setChecked(False)

        if i == -1:
            self.generator_box.edit.setPlainText("")
            self.generator_box.cb_scaling.setCurrentIndex(-1)
            return

        name = self.generators[i]
        default_config = get_generator_defaults(name)

        if name in all_generator_names["bo"]:
            # Patch for BOs that make the low noise prior False by default
            default_config["gp_constructor"]["use_low_noise_prior"] = False
            # Patch for BOs that turn on TuRBO by default
            default_config["turbo_controller"] = "optimize"

        # Patch to only show part of the config
        filtered_config = filter_generator_config(name, default_config)
        self.generator_box.edit.setPlainText(get_yaml_string(filtered_config))

        # Update the docs
        self.window_docs.update_docs(name)

    def _fill_init_table(self):  # make sure self.init_table_actions is set
        for action in self.init_table_actions:
            if action["type"] == "add_curr":
                self.fill_curr_in_init_table(record=False)
            elif action["type"] == "add_rand":
                try:
                    self.add_rand_in_init_table(
                        add_rand_config=action["config"],
                        record=False,
                    )
                except IndexError:  # lower bound is the same as upper bound
                    pass

    def toggle_use_script(self):
        if self.generator_box.check_use_script.isChecked():
            self.generator_box.btn_edit_script.show()
            self.generator_box.edit.setReadOnly(True)
            self.refresh_params_generator()
        else:
            self.generator_box.btn_edit_script.hide()
            self.generator_box.edit.setReadOnly(False)

    def edit_script(self):
        generator = self.generator_box.cb.currentText()
        dlg = BadgerEditScriptDialog(self, generator, self.script, self.script_updated)
        dlg.exec()

    def script_updated(self, text):
        self.script = text
        self.refresh_params_generator()

    def create_env(self):
        env_params = load_config(self.env_box.edit.toPlainText())
        try:
            intf_name = self.configs["interface"][0]
        except KeyError:
            intf_name = None
        configs = {"params": env_params, "interface": [intf_name]}
        env = instantiate_env(self.env, configs)

        return env

    def refresh_params_generator(self):
        if not self.script:
            return

        try:
            tmp = {}
            exec(self.script, tmp)
            try:
                tmp["generate"]  # test if generate function is defined
            except Exception as e:
                QMessageBox.warning(
                    self, "Please define a valid generate function!", str(e)
                )
                return

            env = self.create_env()
            # Get vocs
            try:
                vocs = self._compose_vocs()
            except Exception:
                vocs = None
            # Function generate comes from the script
            params_generator = tmp["generate"](env, vocs)
            self.generator_box.edit.setPlainText(get_yaml_string(params_generator))
        except Exception as e:
            QMessageBox.warning(self, "Invalid script!", str(e))

    def select_env(self, i):
        # Reset the initial table actions and ratio var ranges
        self.init_table_actions = []
        self.ratio_var_ranges = {}

        if i == -1:
            self.env_box.edit.setPlainText("")
            self.env_box.edit_var.clear()
            self.env_box.var_table.update_variables(None)
            self.env_box.edit_obj.clear()
            self.env_box.obj_table.update_objectives(None)
            self.configs = None
            self.env = None
            self.env_box.btn_add_con.setDisabled(True)
            self.env_box.btn_add_sta.setDisabled(True)
            self.env_box.btn_add_var.setDisabled(True)
            self.env_box.btn_lim_vrange.setDisabled(True)
            self.routine = None
            self.env_box.update_stylesheets()
            return

        name = self.envs[i]
        try:
            env, configs = get_env(name)
            self.configs = configs
            self.env = env
            self.env_box.edit_var.clear()
            self.env_box.edit_obj.clear()
            self.env_box.btn_add_con.setDisabled(False)
            self.env_box.btn_add_sta.setDisabled(False)
            self.env_box.btn_add_var.setDisabled(False)
            self.env_box.btn_lim_vrange.setDisabled(False)
            if self.generator_box.check_use_script.isChecked():
                self.refresh_params_generator()
        except Exception:
            self.configs = None
            self.env = None
            self.env_box.cb.setCurrentIndex(-1)
            self.env_box.btn_add_con.setDisabled(True)
            self.env_box.btn_add_sta.setDisabled(True)
            self.env_box.btn_add_var.setDisabled(True)
            self.env_box.btn_lim_vrange.setDisabled(True)
            self.routine = None
            return QMessageBox.critical(self, "Error!", traceback.format_exc())

        self.env_box.edit.setPlainText(get_yaml_string(configs["params"]))

        # Get and save vars to combine with additional vars added on the fly
        vars_env = self.vars_env = configs["variables"]
        vars_combine = [*vars_env]

        self.env_box.check_only_var.blockSignals(True)
        self.env_box.check_only_var.setChecked(False)
        self.env_box.check_only_var.blockSignals(False)
        self.env_box.var_table.update_variables(vars_combine)
        # Auto apply the limited variable ranges if the option is set
        if self.env_box.relative_to_curr.isChecked():
            self.set_vrange(set_all=True)

        # Needed for getting bounds on the fly
        self.env_box.var_table.env_class, self.env_box.var_table.configs = (
            self.add_var()
        )

        _objs_env = configs["observations"]
        objs_env = []
        for name in _objs_env:
            obj = {}
            obj[name] = "MINIMIZE"  # default rule
            objs_env.append(obj)
        self.env_box.check_only_obj.setChecked(False)
        self.env_box.obj_table.update_objectives(objs_env)

        self.env_box.con_table.clear_constraints()
        self.env_box.list_obs.clear()
        self.env_box.fit_content()
        # self.routine = None

        self.env_box.update_stylesheets(env.name)

        # Update the docs
        self.window_env_docs.update_docs(env.name)

    def get_init_table_header(self):
        table = self.env_box.init_table
        header_list = []
        for col in range(table.columnCount()):
            item = table.horizontalHeaderItem(col)
            if item:
                header_list.append(item.text())
            else:
                header_list.append("")  # Handle the case where the header item is None
        return header_list

    def fill_curr_in_init_table(self, record=False):
        env = self.create_env()
        table = self.env_box.init_table
        vname_selected = self.get_init_table_header()
        var_curr = env._get_variables(vname_selected)

        # Iterate through the rows
        for row in range(table.rowCount()):
            # Check if the row is empty
            if np.all(
                [not table.item(row, col).text() for col in range(table.columnCount())]
            ):
                # Fill the row with content_list
                for col, name in enumerate(vname_selected):
                    item = QTableWidgetItem(f"{var_curr[name]:.4g}")
                    table.setItem(row, col, item)
                break  # Stop after filling the first non-empty row

        if record and self.env_box.relative_to_curr.isChecked():
            self.init_table_actions.append({"type": "add_curr"})

    def save_add_rand_config(self, add_rand_config):
        self.add_rand_config = add_rand_config

    def add_rand_in_init_table(self, add_rand_config=None, record=True):
        if add_rand_config is None:
            add_rand_config = self.add_rand_config

        # Get current point
        env = self.create_env()
        vname_selected = self.get_init_table_header()
        var_curr = env._get_variables(vname_selected)

        # get small region around current point to sample
        vocs, _ = self._compose_vocs()
        n_point = add_rand_config["n_points"]
        fraction = add_rand_config["fraction"]
        random_sample_region = get_local_region(var_curr, vocs, fraction=fraction)
        with warnings.catch_warnings(record=True) as caught_warnings:
            random_points = vocs.random_inputs(
                n_point, custom_bounds=random_sample_region
            )

            for warning in caught_warnings:
                # Ignore runtime warnings (usually caused by clip by bounds)
                if warning.category is RuntimeWarning:
                    pass
                else:
                    print(f"Caught user warning: {warning.message}")

        # Add points to the table
        table = self.env_box.init_table
        for row in range(table.rowCount()):
            # Check if the row is empty
            if np.all(
                [not table.item(row, col).text() for col in range(table.columnCount())]
            ):
                # Fill the row with content_list
                try:
                    point = random_points.pop(0)
                    for col, name in enumerate(vname_selected):
                        item = QTableWidgetItem(f"{point[name]:.4g}")
                        table.setItem(row, col, item)
                except IndexError:  # No more points to add
                    break

        if record and self.env_box.relative_to_curr.isChecked():
            self.init_table_actions.append(
                {
                    "type": "add_rand",
                    "config": add_rand_config,
                }
            )

    def show_add_rand_dialog(self):
        dlg = BadgerAddRandomDialog(
            self,
            self.add_rand_in_init_table,
            self.save_add_rand_config,
            self.add_rand_config,
        )
        self.rc_dialog = dlg
        try:
            dlg.exec()
        finally:
            self.rc_dialog = None

    def clear_init_table(self, reset_actions=True):
        table = self.env_box.init_table
        for row in range(table.rowCount()):
            for col in range(table.columnCount()):
                item = table.item(row, col)
                if item:
                    item.setText("")  # Set the cell content to an empty string

        if reset_actions and self.env_box.relative_to_curr.isChecked():
            self.init_table_actions = []  # reset the recorded actions

    def add_row_to_init_table(self):
        table = self.env_box.init_table
        row_position = table.rowCount()
        table.insertRow(row_position)

        for col in range(table.columnCount()):
            item = QTableWidgetItem("")
            table.setItem(row_position, col, item)

    def open_playground(self):
        pass

    def open_generator_docs(self):
        self.window_docs.show()

    def open_environment_docs(self):
        self.window_env_docs.show()

    def add_var(self):
        # TODO: Use a cached env
        env_params = load_config(self.env_box.edit.toPlainText())
        try:
            intf_name = self.configs["interface"][0]
        except KeyError:
            intf_name = None
        configs = {"params": env_params, "interface": [intf_name]}

        return self.env, configs
        # dlg = BadgerVariableDialog(self, self.env, configs, self.add_var_to_list)
        # dlg.exec()

    def limit_variable_ranges(self):
        dlg = BadgerLimitVariableRangeDialog(
            self,
            self.set_vrange,
            self.save_limit_option,
            self.limit_option,
        )
        dlg.exec()

    def set_vrange(self, set_all=False):
        vname_selected = []
        vrange = {}

        for var in self.env_box.var_table.variables:
            name = next(iter(var))
            if set_all or self.env_box.var_table.is_checked(name):
                vname_selected.append(name)
                vrange[name] = var[name]

        env = self.create_env()
        var_curr = env._get_variables(vname_selected)

        option_idx = self.limit_option["limit_option_idx"]
        if option_idx:
            ratio = self.limit_option["ratio_full"]
            for i, name in enumerate(vname_selected):
                hard_bounds = vrange[name]
                delta = 0.5 * ratio * (hard_bounds[1] - hard_bounds[0])
                bounds = [var_curr[name] - delta, var_curr[name] + delta]
                bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()
                vrange[name] = bounds
        else:
            ratio = self.limit_option["ratio_curr"]
            for i, name in enumerate(vname_selected):
                hard_bounds = vrange[name]
                sign = np.sign(var_curr[name])
                bounds = [
                    var_curr[name] * (1 - 0.5 * sign * ratio),
                    var_curr[name] * (1 + 0.5 * sign * ratio),
                ]
                bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()
                vrange[name] = bounds

        self.env_box.var_table.set_bounds(vrange)
        self.clear_init_table(reset_actions=False)  # clear table after changing ranges
        self.update_init_table()  # auto populate if option is set

        # Record the ratio var ranges
        if self.env_box.relative_to_curr.isChecked():
            for vname in vname_selected:
                self.ratio_var_ranges[vname] = copy.deepcopy(self.limit_option)

    def save_limit_option(self, limit_option):
        self.limit_option = limit_option

    def add_var_to_list(self, name, lb, ub):
        # Check if already in the list
        ok = False
        try:
            self.env_box.var_table.bounds[name]
        except KeyError:
            ok = True
        if not ok:
            QMessageBox.warning(
                self, "Variable already exists!", f"Variable {name} already exists!"
            )
            return 1

        self.env_box.add_var(name, lb, ub)
        return 0

    def update_init_table(self):
        selected = self.env_box.var_table.selected
        variable_names = [v for v in selected if selected[v]]
        update_init_data_table(self.env_box.init_table, variable_names)

        if not self.env_box.relative_to_curr.isChecked():
            return

        # Auto populate the initial table based on recorded actions
        if not self.init_table_actions:
            self.init_table_actions = [
                {"type": "add_curr"},
                {"type": "add_rand", "config": self.add_rand_config},
            ]
        self.clear_init_table(reset_actions=False)
        self._fill_init_table()

    def calc_auto_bounds(self):
        vname_selected = []
        vrange = {}

        for var in self.env_box.var_table.variables:
            name = next(iter(var))
            vname_selected.append(name)
            vrange[name] = var[name]

        env = self.create_env()
        var_curr = env._get_variables(vname_selected)

        for name in vname_selected:
            try:
                limit_option = self.ratio_var_ranges[name]
            except KeyError:
                limit_option = self.limit_option

            option_idx = limit_option["limit_option_idx"]
            if option_idx:
                ratio = limit_option["ratio_full"]
                hard_bounds = vrange[name]
                delta = 0.5 * ratio * (hard_bounds[1] - hard_bounds[0])
                bounds = [var_curr[name] - delta, var_curr[name] + delta]
                bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()
                vrange[name] = bounds
            else:
                ratio = limit_option["ratio_curr"]
                hard_bounds = vrange[name]
                sign = np.sign(var_curr[name])
                bounds = [
                    var_curr[name] * (1 - 0.5 * sign * ratio),
                    var_curr[name] * (1 + 0.5 * sign * ratio),
                ]
                bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()
                vrange[name] = bounds

        return vrange

    def toggle_relative_to_curr(self, checked, refresh=True):
        if checked:
            self.env_box.switch_var_panel_style(True)

            if refresh and self.env_box.var_table.selected:
                bounds = self.calc_auto_bounds()
                self.env_box.var_table.set_bounds(bounds)
                self.clear_init_table(reset_actions=False)
                # Auto populate the initial table
                self.try_populate_init_table()

            self.env_box.var_table.lock_bounds()
            self.env_box.init_table.set_uneditable()
        else:
            self.env_box.switch_var_panel_style(False)

            self.env_box.var_table.unlock_bounds()
            self.env_box.init_table.set_editable()

    def try_populate_init_table(self):
        if (
            self.env_box.relative_to_curr.isChecked()
            and self.env_box.var_table.selected
        ):
            self.update_init_table()

    def handle_pv_added(self):
        if self.env_box.relative_to_curr.isChecked():
            self.set_vrange()

    def add_constraint(self, name=None, relation=0, threshold=0, critical=False):
        if self.configs is None:
            return

        options = self.configs["observations"]
        self.env_box.con_table.add_constraint(
            options,
            name,
            relation,
            threshold,
            critical,
        )

    def add_state(self, name=None):
        if self.configs is None:
            return

        var_names = [next(iter(d)) for d in self.configs["variables"]]
        options = self.configs["observations"] + var_names
        item = QListWidgetItem(self.env_box.list_obs)
        sta_item = state_item(
            options,
            lambda: self.env_box.list_obs.takeItem(self.env_box.list_obs.row(item)),
            name,
        )
        item.setSizeHint(sta_item.sizeHint())
        self.env_box.list_obs.addItem(item)
        self.env_box.list_obs.setItemWidget(item, sta_item)
        self.env_box.fit_content()

    def _compose_vocs(self) -> (VOCS, list[str]):
        # Compose the VOCS settings
        variables = self.env_box.var_table.export_variables()
        objectives = self.env_box.obj_table.export_objectives()

        constraints = {}
        critical_constraints = []
        for (
            con_name,
            relation,
            threshold,
            critical,
            _,
        ) in self.env_box.con_table.export_constraints():
            constraints[con_name] = [relation, threshold]
            if critical:
                critical_constraints.append(con_name)

        observables = []
        for i in range(self.env_box.list_obs.count()):
            item = self.env_box.list_obs.item(i)
            item_widget = self.env_box.list_obs.itemWidget(item)
            obs_name = item_widget.cb_sta.currentText()
            observables.append(obs_name)

        vocs = VOCS(
            variables=variables,
            objectives=objectives,
            constraints=constraints,
            constants={},
            observables=observables,
        )

        return vocs, critical_constraints

    def _compose_routine(self) -> Routine:
        # Compose the routine
        name = self.edit_save.text() or self.edit_save.placeholderText()
        description = self.edit_descr.toPlainText()

        if self.generator_box.cb.currentIndex() == -1:
            raise BadgerRoutineError("no generator selected")
        if self.env_box.cb.currentIndex() == -1:
            raise BadgerRoutineError("no environment selected")

        generator_name = self.generators[self.generator_box.cb.currentIndex()]
        env_name = self.envs[self.env_box.cb.currentIndex()]
        generator_params = load_config(self.generator_box.edit.toPlainText())
        if generator_name in all_generator_names["bo"]:
            # Patch the BO generators to make sure use_low_noise_prior is False
            if "gp_constructor" not in generator_params:
                generator_params["gp_constructor"] = {
                    "name": "standard",  # have to add name too for pydantic validation
                    "use_low_noise_prior": False,
                }
            # or else we use whatever specified by the users

            # Patch the BO generators to turn on TuRBO by default
            if "turbo_controller" not in generator_params:
                generator_params["turbo_controller"] = "optimize"
        env_params = load_config(self.env_box.edit.toPlainText())

        # VOCS
        vocs, critical_constraints = self._compose_vocs()

        # Initial points
        init_points_df = pd.DataFrame.from_dict(
            get_table_content_as_dict(self.env_box.init_table)
        )
        init_points_df = init_points_df.replace("", pd.NA)
        init_points_df = init_points_df.dropna(subset=init_points_df.columns, how="all")
        contains_na = init_points_df.isna().any().any()
        if contains_na:
            raise BadgerRoutineError(
                "Initial points are not valid, please fill in the missing values"
            )

        # Script that generates generator params
        if self.generator_box.check_use_script.isChecked():
            script = self.script
        else:
            script = None

        # Relative to current params
        if self.env_box.relative_to_curr.isChecked():
            relative_to_current = True
            vrange_limit_options = self.ratio_var_ranges
            initial_point_actions = self.init_table_actions
        else:
            relative_to_current = False
            vrange_limit_options = None
            initial_point_actions = None

        with warnings.catch_warnings(record=True) as caught_warnings:
            routine = Routine(
                # Metadata
                badger_version=get_badger_version(),
                xopt_version=get_xopt_version(),
                # Xopt part
                vocs=vocs,
                generator={"name": generator_name} | generator_params,
                # Badger part
                name=name,
                description=description,
                environment={"name": env_name} | env_params,
                initial_points=init_points_df.astype("double"),
                critical_constraint_names=critical_constraints,
                tags=None,
                script=script,
                relative_to_current=relative_to_current,
                vrange_limit_options=vrange_limit_options,
                initial_point_actions=initial_point_actions,
                additional_variables=self.env_box.var_table.addtl_vars,
            )

            # Check if any user warnings were caught
            for warning in caught_warnings:
                if isinstance(warning.category, UserWarning):
                    pass
                else:
                    print(f"Caught user warning: {warning.message}")

            return routine

    def review(self):
        try:
            routine = self._compose_routine()
        except:
            return QMessageBox.critical(
                self, "Invalid routine!", traceback.format_exc()
            )

        dlg = BadgerReviewDialog(self, routine)
        dlg.exec()

    def update_description(self):
        routine = self.routine
        routine.description = self.edit_descr.toPlainText()
        try:
            update_routine(routine)
            # Notify routine list to update
            self.sig_updated.emit(routine.name, routine.description)
            QMessageBox.information(
                self,
                "Update success!",
                f"Routine {self.routine.name} description was updated!",
            )
        except Exception:
            return QMessageBox.critical(self, "Update failed!", traceback.format_exc())

    def save(self):
        try:
            routine = self._compose_routine()
        except ValidationError as e:
            error_message = "".join(
                [error["msg"] + "\n\n" for error in e.errors()]
            ).strip()
            details = traceback.format_exc()
            dialog = ExpandableMessageBox(
                title="Error!", text=error_message, detailedText=details, parent=self
            )
            dialog.setIcon(QMessageBox.Critical)
            dialog.exec_()
            return

        try:
            if self.routine:
                keys_to_exclude = ["data", "id", "name", "description"]
                old_dict = json.loads(self.routine.json())
                old_dict = {
                    k: v for k, v in old_dict.items() if k not in keys_to_exclude
                }
                new_dict = json.loads(routine.json())
                new_dict = {
                    k: v for k, v in new_dict.items() if k not in keys_to_exclude
                }
                runs = get_runs_by_routine(self.routine.id)
                if len(runs) == 0 or old_dict == new_dict:
                    routine.id = self.routine.id
                    update_routine(routine)
                else:
                    save_routine(routine)
            else:
                save_routine(routine)
        except sqlite3.IntegrityError:
            return QMessageBox.critical(
                self,
                "Error!",
                f"Routine {routine.name} already existed in the database! Please "
                f"choose another name.",
            )

        return 0



def test_routines_same_name():
    from badger.db import save_routine, remove_routine
    from badger.tests.utils import create_routine, fix_db_path_issue

    fix_db_path_issue()

    routine1 = create_routine()
    routine2 = create_routine()
    save_routine(routine1)
    save_routine(routine2)

    assert routine1.id != routine2.id

    remove_routine(routine1.id)
    remove_routine(routine2.id)


def test_modify_routine_no_runs(qtbot):
    from badger.gui.default.components.routine_page import BadgerRoutinePage
    from badger.db import list_routine, remove_routine, load_routine

    window = BadgerRoutinePage()
    qtbot.addWidget(window)
    window.env_box.relative_to_curr.setChecked(False)
    qtbot.keyClicks(window.generator_box.cb, "expected_improvement")
    qtbot.keyClicks(window.env_box.cb, "test")

    window.env_box.var_table.cellWidget(0, 0).setChecked(True)
    assert window.env_box.var_table.export_variables() == {"x0": [-1, 1]}

    window.env_box.obj_table.cellWidget(0, 0).setChecked(True)
    assert window.env_box.obj_table.export_objectives() == {"f": "MINIMIZE"}

    window.save()
    id = list_routine()[0][0]
    routine, _ = load_routine(id)
    window.refresh_ui(routine)

    index = window.generator_box.cb.findText("extremum_seeking")
    window.generator_box.cb.setCurrentIndex(index)
    window.save()

    ids = list_routine()[0]
    assert len(ids) == 1
    for id in ids:
        remove_routine(id)


# TODO: write test for modifying name of routine with runs
def test_modify_routine_name(qtbot):
    from badger.gui.default.components.routine_page import BadgerRoutinePage
    from badger.db import list_routine, remove_routine, load_routine, save_run

    window = BadgerRoutinePage()
    qtbot.addWidget(window)
    window.env_box.relative_to_curr.setChecked(False)
    qtbot.keyClicks(window.generator_box.cb, "expected_improvement")
    qtbot.keyClicks(window.env_box.cb, "test")
    window.env_box.var_table.cellWidget(0, 0).setChecked(True)
    window.env_box.obj_table.cellWidget(0, 0).setChecked(True)

    window.save()
    id = list_routine()[0][0]
    routine, _ = load_routine(id)
    window.refresh_ui(routine)

    run = {
        "filename": "BadgerOpt-2024-09-10-155408.yaml",
        "routine": routine,
        "data": {
            "f": [0.051416649999999994],
            "timestamp": [1729263544.34316],
            # 'x0': [-0.1711],
            # 'xopt_error': [False],
            # 'xopt_runtime': [0.0007418617606163025]
        },
    }
    save_run(run)

    window.edit_save.setText("new_name")
    window.save()

    ids = list_routine()[0]
    assert len(ids) == 1
    for id in ids:
        remove_routine(id)


# TODO: write test for modifying algorithm of routine with runs
def test_modify_routine_algorithm(qtbot):
    from badger.gui.default.components.routine_page import BadgerRoutinePage
    from badger.db import list_routine, remove_routine, load_routine, save_run

    window = BadgerRoutinePage()
    qtbot.addWidget(window)
    window.env_box.relative_to_curr.setChecked(False)
    qtbot.keyClicks(window.generator_box.cb, "expected_improvement")
    qtbot.keyClicks(window.env_box.cb, "test")
    window.env_box.var_table.cellWidget(0, 0).setChecked(True)
    window.env_box.obj_table.cellWidget(0, 0).setChecked(True)

    window.save()
    id = list_routine()[0][0]
    routine, _ = load_routine(id)
    window.refresh_ui(routine)

    run = {
        "filename": "BadgerOpt-2024-09-10-155408.yaml",
        "routine": routine,
        "data": {
            "f": [0.051416649999999994],
            "timestamp": [1729263544.34316],
            # 'x0': [-0.1711],
            # 'xopt_error': [False],
            # 'xopt_runtime': [0.0007418617606163025]
        },
    }
    save_run(run)

    index = window.generator_box.cb.findText("extremum_seeking")
    window.generator_box.cb.setCurrentIndex(index)
    window.save()

    ids = list_routine()[0]
    assert len(ids) == 2
    for id in ids:
        remove_routine(id)



import logging
import os
import sys
import time
import signal

from pandas import DataFrame

from badger.utils import curr_ts
from badger.core import run_routine as run
from badger.routine import Routine
from badger.settings import init_settings
from badger.errors import BadgerRunTerminated

logger = logging.getLogger(__name__)


def run_n_archive(
    routine: Routine, yes=False, save=False, verbose=2, sleep=0, flush_prompt=False
):
    try:
        from badger.archive import archive_run
    except Exception as e:
        logger.error(e)
        return

    # Store system states and other stuff
    storage = {
        "states": None,
        "ts_last_dump": None,
        "paused": False,
    }

    def handler(*args):
        if storage["paused"]:
            print("")  # start a new line
            if flush_prompt:  # erase the last prompt
                sys.stdout.write("\033[F")
            raise BadgerRunTerminated
        storage["paused"] = True

    signal.signal(signal.SIGINT, handler)

    def check_run_status():
        return 0

    def before_evaluate(candidates: DataFrame):
        if storage["paused"]:
            res = input(
                "Optimization paused. Press Enter to resume or Ctrl/Cmd + C to terminate: "
            )
            while res != "":
                if flush_prompt:
                    sys.stdout.write("\033[F")
                res = input(
                    f"Invalid choice: {res}. Please press Enter to resume or Ctrl/Cmd + C to terminate: "
                )
            if flush_prompt:
                sys.stdout.write("\033[F")
        storage["paused"] = False

    def after_evaluate(data: DataFrame):
        # vars: ndarray
        # obses: ndarray
        # cons: ndarray
        # stas: list
        ts = curr_ts()
        ts_float = ts.timestamp()
        config = init_settings()
        # Try dump the run data and interface log to the disk
        dump_period = float(config.read_value("BADGER_DATA_DUMP_PERIOD"))
        ts_last_dump = storage["ts_last_dump"]
        if (ts_last_dump is None) or (ts_float - ts_last_dump > dump_period):
            storage["ts_last_dump"] = ts_float
            _run = archive_run(routine, storage["states"])
            # Try dump the interface logs
            try:
                path = _run["path"]
                filename = _run["filename"][:-4] + "pickle"
                routine.environment.interface.dump_recording(
                    os.path.join(path, filename)
                )
            except Exception:
                pass

        # take a break to let the outside signal to change the status
        time.sleep(sleep)

    def states_ready(states):
        storage["states"] = states

    try:
        run(
            routine,
            active_callback=check_run_status,
            generate_callback=before_evaluate,
            evaluate_callback=after_evaluate,
            states_callback=states_ready,
        )
    except BadgerRunTerminated as e:
        logger.info(e)
    except Exception as e:
        logger.error(e)

    # Save the run when at least one solution has been evaluated
    if len(routine.data):
        _run = archive_run(routine, storage["states"])
        # Try dump the interface logs
        try:
            path = _run["path"]
            filename = _run["filename"][:-4] + "pickle"
            routine.environment.interface.stop_recording(os.path.join(path, filename))
        except Exception:
            pass


def run_routine(args):
    print(
        "This command is deprecated.\n"
        "Please use 'badger -g' to launch the Badger GUI "
        "and run an optimization."
    )
    return

    # try:
    #     from ..factory import get_algo, get_env
    # except Exception as e:
    #     logger.error(e)
    #     return

    # try:
    #     # Get env params
    #     _, configs_env = get_env(args.env)

    #     # Get algo params
    #     _, configs_algo = get_algo(args.algo)

    #     # Normalize the algo and env params
    #     params_env = load_config(args.env_params)
    #     params_algo = load_config(args.algo_params)
    # except Exception as e:
    #     logger.error(e)
    #     return
    # params_env = merge_params(configs_env['params'], params_env)
    # params_algo = merge_params(configs_algo['params'], params_algo)

    # # Load routine configs
    # try:
    #     configs_routine = load_config(args.config)
    # except Exception as e:
    #     logger.error(e)
    #     return

    # # Compose the routine
    # routine = {
    #     'name': args.save or generate_slug(2),
    #     'algo': args.algo,
    #     'env': args.env,
    #     'algo_params': params_algo,
    #     'env_params': params_env,
    #     # env_vranges is an additional info for the normalization
    #     # Will be removed after the normalization
    #     'env_vranges': config_list_to_dict(configs_env['variables']),
    #     'config': configs_routine,
    # }

    # run_n_archive(routine, args.yes, args.save, args.verbose)



import subprocess
from importlib import metadata


def capture(command):
    proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate()
    out = out.decode("utf-8")
    err = err.decode("utf-8")
    return out, err, proc.returncode


def test_cli_main():
    command = ["badger"]
    out, err, exitcode = capture(command)

    assert exitcode == 0

    # Check output lines
    outlines = out.splitlines()
    assert len(outlines) == 7

    # Check name
    assert outlines[0] == "name: Badger the optimizer"

    # Check version
    version = metadata.version("badger-opt")
    try:  # yaml encoding number-like string differently
        _ = float(version)
        assert outlines[1] == f"version: '{version}'"
    except ValueError:
        assert outlines[1] == f"version: {version}"


def test_list_algo():
    from badger.factory import ALGO_EXCLUDED

    command = ["badger", "generator"]
    out, err, exitcode = capture(command)

    assert exitcode == 0

    # Check output lines
    outlines = out.splitlines()
    for algo in ["expected_improvement", "neldermead", "rcds"]:
        if algo not in ALGO_EXCLUDED:
            assert f"- {algo}" in outlines


# def test_cli_run(mock_config_root):
#     command = ['badger', 'run', '-a', 'upper_confidence_bound', '-ap',
#                '{max_evaluations: 10}',  '-e', 'silly', '-c',
#                os.path.join(mock_config_root, 'test.yaml'), '-y']
#     out, err, exitcode = capture(command)

#     assert exitcode == 0

#     # Check output lines
#     outlines = out.splitlines()
#     assert len(outlines) == 15

#     # Check table header
#     assert outlines[1] == '|    iter    |     l2     ' \
#         + '|     q1     |     q2     |'



from badger.settings import init_settings
from badger.utils import get_value_or_none
from badger.errors import (
    BadgerConfigError,
    BadgerInvalidPluginError,
    BadgerInvalidDocsError,
    BadgerPluginNotFoundError,
)
import sys
import os
import importlib
import yaml
from xopt.generators import generators, get_generator_defaults

import logging

logger = logging.getLogger(__name__)

LOAD_LOCAL_ALGO = False
ALGO_EXCLUDED = [
    "bayesian_exploration",
    "cnsga",
    "mggpo",
    "time_dependent_upper_confidence_bound",
    "multi_fidelity",
    "nsga2",
]

# Check badger plugin root
config_singleton = init_settings()
BADGER_PLUGIN_ROOT = config_singleton.read_value("BADGER_PLUGIN_ROOT")
if BADGER_PLUGIN_ROOT is None:
    raise BadgerConfigError("Please set the BADGER_PLUGIN_ROOT env var!")
elif not os.path.exists(BADGER_PLUGIN_ROOT):
    raise BadgerConfigError(
        f"The badger plugin root {BADGER_PLUGIN_ROOT} does not exist!"
    )
else:
    module_file = os.path.join(BADGER_PLUGIN_ROOT, "__init__.py")
    if not os.path.exists(module_file):
        with open(module_file, "w") as f:
            pass
sys.path.append(BADGER_PLUGIN_ROOT)


def scan_plugins(root):
    factory = {}

    # Do not scan local generators if option disabled
    if LOAD_LOCAL_ALGO:
        ptype_list = ["generator", "interface", "environment"]
    else:
        ptype_list = ["interface", "environment"]
        factory["generator"] = {}

    for ptype in ptype_list:
        factory[ptype] = {}

        proot = os.path.join(root, f"{ptype}s")

        try:
            plugins = [
                fname
                for fname in os.listdir(proot)
                if os.path.exists(os.path.join(proot, fname, "__init__.py"))
            ]
        except:
            plugins = []

        for pname in plugins:
            # TODO: Also load the configs here
            # So that list plugins can access the metadata of the plugins
            factory[ptype][pname] = None

    return factory


def load_plugin(root, pname, ptype):
    assert ptype in [
        "generator",
        "interface",
        "environment",
    ], f"Invalid plugin type {ptype}"

    proot = os.path.join(root, f"{ptype}s")

    # Load the params in the configs
    configs = None
    with open(os.path.join(proot, pname, "configs.yaml"), "r") as f:
        try:
            configs = yaml.safe_load(f)
        except yaml.YAMLError:
            raise BadgerInvalidPluginError(
                f"Error loading plugin {ptype} {pname}: invalid config"
            )

    # Load module
    try:
        module = importlib.import_module(f"{ptype}s.{pname}")
    except ImportError as e:
        _e = BadgerInvalidPluginError(
            f"{ptype} {pname} is not available due to missing dependencies: {e}"
        )
        _e.configs = configs  # attach information to the exception
        raise _e

    if ptype == "generator":
        plugin = [module.optimize, configs]
    elif ptype == "interface":
        params = module.Interface.model_json_schema()["properties"]
        params = {
            name: get_value_or_none(info, "default") for name, info in params.items()
        }
        configs["params"] = params
        plugin = [module.Interface, configs]
    elif ptype == "environment":
        vars = module.Environment.variables
        obses = module.Environment.observables
        params = module.Environment.model_json_schema()["properties"]
        params = {
            name: get_value_or_none(info, "default")
            for name, info in params.items()
            if name != "interface"
        }
        # Get vranges by creating an env instance
        try:
            intf_name = configs["interface"][0]
            Interface, _ = get_intf(intf_name)
            intf = Interface()
        except KeyError:
            intf = None
        except Exception as e:
            logger.warning(e)
            intf = None
        env = module.Environment(interface=intf, params=configs)
        var_bounds = env._get_bounds(vars)

        vars_info = []
        for var in vars:
            var_info = {}
            var_info[var] = var_bounds[var]
            vars_info.append(var_info)

        configs["params"] = params
        configs["variables"] = vars_info
        configs["observations"] = obses
        plugin = [module.Environment, configs]
    else:  # TODO: raise an exception here instead?
        return [None, None]

    BADGER_FACTORY[ptype][pname] = plugin

    return plugin


def load_docs(root, pname, ptype):
    assert ptype in [
        "generator",
        "interface",
        "environment",
    ], f"Invalid plugin type {ptype}"

    proot = os.path.join(root, f"{ptype}s")

    # Load the readme
    readme = None
    try:
        with open(os.path.join(proot, pname, "README.md"), "r") as f:
            readme = f.read()
        return readme
    except:
        raise BadgerInvalidDocsError(
            f"Error loading docs for {ptype} {pname}: docs not found"
        )


def get_plug(root, name, ptype):
    try:
        plug = BADGER_FACTORY[ptype][name]
        if plug is None:  # lazy loading
            plug = load_plugin(root, name, ptype)
            BADGER_FACTORY[ptype][name] = plug
        # Prevent accidentially modifying default configs
        plug = [plug[0], plug[1].copy()]
    except KeyError:
        raise BadgerPluginNotFoundError(
            f"Error loading plugin {ptype} {name}: plugin not found"
        )

    return plug


def scan_extensions(root):
    extensions = {}

    return extensions


def get_generator_docs(name):
    return generators[name].__doc__


def get_env_docs(name):
    return load_docs(BADGER_PLUGIN_ROOT, name, "environment")


def get_intf(name):
    return get_plug(BADGER_PLUGIN_ROOT, name, "interface")


def get_env(name):
    return get_plug(BADGER_PLUGIN_ROOT, name, "environment")


def list_generators():
    try:
        from xopt.generators import try_load_all_generators

        try_load_all_generators()
    except ImportError:  # this API changed somehow
        pass  # there is nothing we can do...
    generator_names = list(generators.keys())
    # Filter the names
    generator_names = [n for n in generator_names if n not in ALGO_EXCLUDED]
    return sorted(generator_names)


get_generator = get_generator_defaults


def list_intf():
    return sorted(BADGER_FACTORY["interface"])


def list_env():
    return sorted(BADGER_FACTORY["environment"])


BADGER_FACTORY = scan_plugins(BADGER_PLUGIN_ROOT)
BADGER_EXTENSIONS = scan_extensions(BADGER_PLUGIN_ROOT)



import json
from copy import deepcopy
from typing import Any, List, Optional

import numpy as np
import pandas as pd
from pandas import DataFrame
from pydantic import (
    ConfigDict,
    Field,
    field_validator,
    model_validator,
    SerializeAsAny,
    ValidationInfo,
)
from xopt import Evaluator, VOCS, Xopt
from xopt.generators import get_generator
from xopt.utils import get_local_region
from xopt.generators.sequential import SequentialGenerator
from badger.utils import curr_ts
from badger.environment import Environment, instantiate_env


class Routine(Xopt):
    id: Optional[str] = Field(None)
    name: str
    description: Optional[str] = Field(None)
    environment: SerializeAsAny[Environment]
    initial_points: Optional[DataFrame] = Field(None)
    critical_constraint_names: Optional[List[str]] = Field([])
    tags: Optional[List] = Field(None)
    script: Optional[str] = Field(None)
    # Store relative to current params
    relative_to_current: Optional[bool] = Field(False)
    vrange_limit_options: Optional[dict] = Field(None)
    vrange_hard_limit: Optional[dict] = Field({})  # override hard limits
    initial_point_actions: Optional[List] = Field(None)
    additional_variables: Optional[List[str]] = Field([])
    formulas: Optional[dict[str, dict[str, Any]]] = Field({})
    # Other meta data
    badger_version: Optional[str] = Field(None)
    xopt_version: Optional[str] = Field(None)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="before")
    @classmethod
    def validate_model(cls, data: Any):
        from badger.factory import get_env

        if isinstance(data, dict):
            # validate vocs
            if isinstance(data["vocs"], dict):
                data["vocs"] = VOCS(**data["vocs"])

            # validate generator
            if isinstance(data["generator"], dict):
                name = data["generator"].pop("name")
                generator_class = get_generator(name)
                data["generator"] = generator_class.model_validate(
                    {**data["generator"], "vocs": data["vocs"]}
                )

            elif isinstance(data["generator"], str):
                generator_class = get_generator(data["generator"])

                data["generator"] = generator_class.model_validate(
                    {"vocs": data["vocs"]}
                )

            if isinstance(data["generator"], SequentialGenerator):
                data["generator"].is_active = False

            # validate data (if it exists
            if "data" in data:
                if isinstance(data["data"], dict):
                    try:
                        data["data"] = pd.DataFrame(data["data"])
                    except IndexError:
                        data["data"] = pd.DataFrame(data["data"], index=[0])

                    data["data"].index = data["data"].index.astype(int)
                    data["data"].sort_index(inplace=True)

                    # Add data one row at a time to avoid generator issues
                    if isinstance(data["generator"], SequentialGenerator):
                        data["generator"].set_data(data["data"])
                    else:
                        data["generator"].add_data(data["data"])

            # instantiate env
            if isinstance(data["environment"], dict):
                # TODO: Actually we need this interface info, but
                # should be put somewhere else (in parallel with env?)
                try:
                    del data["environment"]["interface"]
                except KeyError:  # no interface at all, which is good
                    pass
                name = data["environment"].pop("name")
                env_class, configs_env = get_env(name)
                configs_env["params"] |= data["environment"]
                data["environment"] = instantiate_env(env_class, configs_env)
            else:  # should be an instantiated env already
                pass

            # create evaluator
            env = data["environment"]

            def evaluate_point(point: dict):
                # sanitize inputs
                point = pd.Series(point).explode().to_dict()
                env._set_variables(point)
                obs = env._get_observables(data["vocs"].output_names)

                ts = curr_ts()
                obs["timestamp"] = ts.timestamp()

                return obs

            data["evaluator"] = Evaluator(function=evaluate_point)

        return data

    @field_validator("initial_points", mode="before")
    def validate_data(cls, v, info: ValidationInfo):
        if isinstance(v, dict):
            try:
                v = pd.DataFrame(v)
            except IndexError:
                v = pd.DataFrame(v, index=[0])

        return v

    @property
    def sorted_data(self):
        data_copy = deepcopy(self.data)
        if data_copy is not None:
            data_copy.index = data_copy.index.astype(int)
            data_copy.sort_index(inplace=True)

        return data_copy

    def json(self, **kwargs) -> str:
        """Handle custom serialization of environment"""

        result = super().json(**kwargs)
        dict_result = json.loads(result)

        # Remove extra fields
        fields_to_be_removed = [
            "dump_file",
            "evaluator",
            "max_evaluations",
            "serialize_inline",
            "serialize_torch",
            "strict",
        ]
        for field in fields_to_be_removed:
            dict_result.pop(field, None)

        dict_result["environment"] = {"name": self.environment.name} | dict_result[
            "environment"
        ]
        try:
            dict_result["environment"]["interface"] = {
                "name": self.environment.interface.name
            } | dict_result["environment"]["interface"]
        except KeyError:
            pass
        except AttributeError:
            pass

        return json.dumps(dict_result)


def calculate_variable_bounds(limit_options, vocs, env):
    vnames = vocs.variable_names
    var_curr = env._get_variables(vnames)
    var_range = env._get_bounds(vnames)

    variables_updated = {}
    for name in vnames:
        try:
            limit_option = limit_options[name]
        except KeyError:
            continue

        option_idx = limit_option["limit_option_idx"]
        if option_idx:
            ratio = limit_option["ratio_full"]
            hard_bounds = var_range[name]
            delta = 0.5 * ratio * (hard_bounds[1] - hard_bounds[0])
            bounds = [var_curr[name] - delta, var_curr[name] + delta]
            bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()
            variables_updated[name] = bounds
        else:
            ratio = limit_option["ratio_curr"]
            hard_bounds = var_range[name]
            sign = np.sign(var_curr[name])
            bounds = [
                var_curr[name] * (1 - 0.5 * sign * ratio),
                var_curr[name] * (1 + 0.5 * sign * ratio),
            ]
            bounds = np.clip(bounds, hard_bounds[0], hard_bounds[1]).tolist()
            variables_updated[name] = bounds

    return variables_updated


def calculate_initial_points(init_actions, vocs, env):
    vnames = vocs.variable_names
    init_points = {k: [] for k in vnames}

    for action in init_actions:
        if action["type"] == "add_curr":
            var_curr = env._get_variables(vnames)
            for name in vnames:
                init_points[name].append(var_curr[name])
        elif action["type"] == "add_rand":
            var_curr = env._get_variables(vnames)
            n_point = action["config"]["n_points"]
            fraction = action["config"]["fraction"]
            random_sample_region = get_local_region(var_curr, vocs, fraction=fraction)
            random_points = vocs.random_inputs(
                n_point, custom_bounds=random_sample_region
            )
            for point in random_points:
                for name in vnames:
                    init_points[name].append(point[name])

    return init_points



import json
from copy import deepcopy

from xopt.generators import get_generator, get_generator_defaults
from xopt.resources.testing import TEST_VOCS_BASE


class TestFactory:
    def test_generator_generation(self):
        from badger.factory import list_generators

        generator_names = list_generators()

        # test serializing and loading from objects
        for name in generator_names:
            gen_config = get_generator_defaults(name)
            gen_class = get_generator(name)

            if name in ["mobo"]:
                test_vocs = deepcopy(TEST_VOCS_BASE)
                test_vocs.objectives = {"y1": "MINIMIZE", "y2": "MINIMIZE"}
                gen_config["reference_point"] = {"y1": 10.0, "y2": 1.0}
                json.dumps(gen_config)

                gen_class(vocs=test_vocs, **gen_config)
            elif name in ["multi_fidelity", "extremum_seeking", "neldermead", "rcds"]:
                test_vocs = deepcopy(TEST_VOCS_BASE)
                test_vocs.constraints = {}
                json.dumps(gen_config)

                gen_class(vocs=test_vocs, **gen_config)
            elif name in ["bayesian_exploration"]:
                test_vocs = deepcopy(TEST_VOCS_BASE)
                test_vocs.objectives = {}
                test_vocs.observables = ["f"]
                json.dumps(gen_config)

                gen_class(vocs=test_vocs, **gen_config)
            else:
                json.dumps(gen_config)
                gen_class(vocs=TEST_VOCS_BASE, **gen_config)



import multiprocessing
import warnings
from unittest.mock import patch

import numpy as np
import pytest
from PyQt5.QtCore import QEventLoop, QPointF, Qt, QTimer
from PyQt5.QtGui import QMouseEvent
from PyQt5.QtTest import QSignalSpy
from PyQt5.QtWidgets import QApplication, QMessageBox


class TestRunMonitor:
    @pytest.fixture(scope="session")
    def init_multiprocessing(self):
        multiprocessing.set_start_method("fork", force=True)

    @pytest.fixture
    def process_manager(self):
        from badger.gui.default.components.create_process import CreateProcess
        from badger.gui.default.components.process_manager import ProcessManager

        process_manager = ProcessManager()
        process_builder = CreateProcess()
        process_builder.subprocess_prepared.connect(process_manager.add_to_queue)
        process_builder.create_subprocess()
        yield process_manager

        process_manager.close_proccesses()

    @pytest.fixture
    def monitor(self, process_manager, init_multiprocessing):
        from badger.archive import save_tmp_run
        from badger.gui.default.components.run_monitor import BadgerOptMonitor
        from badger.tests.utils import create_routine

        routine = create_routine()
        tmp_filename = save_tmp_run(routine)
        monitor = BadgerOptMonitor(process_manager)
        monitor.routine_filename = tmp_filename
        monitor.testing = True
        monitor.routine = routine

        return monitor

    @pytest.fixture
    def home_page(self, process_manager):
        from badger.archive import save_tmp_run
        from badger.gui.acr.pages.home_page import BadgerHomePage
        from badger.tests.utils import create_routine

        routine = create_routine()
        tmp_filename = save_tmp_run(routine)
        home = BadgerHomePage(process_manager)
        home.current_routine = routine

        monitor = home.run_monitor
        monitor.routine_filename = tmp_filename
        monitor.testing = True
        monitor.init_plots(routine)

        return home

    @pytest.fixture
    def home_page_critical(self, process_manager):
        from badger.archive import save_tmp_run
        from badger.gui.acr.pages.home_page import BadgerHomePage
        from badger.tests.utils import create_routine_critical

        routine = create_routine_critical()
        tmp_filename = save_tmp_run(routine)
        home = BadgerHomePage(process_manager)
        home.current_routine = routine

        monitor = home.run_monitor
        monitor.routine_filename = tmp_filename
        monitor.testing = True
        monitor.init_plots(routine)

        return home

    def add_data(self, monitor):
        monitor.routine.random_evaluate(10)
        monitor.init_plots(monitor.routine)

        assert len(monitor.routine.data) == 10

    def test_run_monitor(self, process_manager):
        from badger.gui.default.components.run_monitor import BadgerOptMonitor
        from badger.tests.utils import create_routine

        monitor = BadgerOptMonitor(process_manager)
        monitor.testing = True
        # qtbot.addWidget(monitor)

        routine = create_routine()
        monitor.init_plots(routine)

        # add some data
        monitor.routine.step()
        assert len(monitor.routine.data) == 1

        # test updating plots
        monitor.update_curves()
        assert set(monitor.curves_variable.keys()) == {"x0", "x1", "x2", "x3"}
        assert set(monitor.curves_objective.keys()) == {"f"}
        assert set(monitor.curves_constraint.keys()) == {"c"}

        # set up run monitor and test it
        monitor.init_routine_runner()
        monitor.routine_runner.set_termination_condition({"tc_idx": 0, "max_eval": 2})
        spy = QSignalSpy(monitor.routine_runner.signals.progress)
        assert spy.isValid()
        monitor.start()

    def test_routine_identity(self, home_page):
        monitor = home_page.run_monitor
        monitor.init_routine_runner()

        assert monitor.routine_runner.routine == monitor.routine

    def test_plotting(self, qtbot, monitor):
        self.add_data(monitor)
        monitor.update_curves()

        monitor.plot_x_axis = 1
        monitor.update_curves()

        monitor.plot_x_axis = 0
        monitor.x_plot_relative = 1
        monitor.update_curves()

        monitor.plot_x_axis = 1
        monitor.x_plot_relative = 1
        monitor.x_plot_y_axis = 1
        monitor.update_curves()

    def test_click_graph(self, qtbot, monitor, mocker):
        self.add_data(monitor)
        sig_inspect_spy = QSignalSpy(monitor.sig_inspect)
        monitor.plot_x_axis = True

        mock_event = mocker.MagicMock(spec=QMouseEvent)
        mock_event._scenePos = QPointF(350, 240)

        orginal_value = monitor.inspector_variable.value()
        monitor.on_mouse_click(mock_event)
        new_variable_value = monitor.inspector_variable.value()

        assert new_variable_value != orginal_value
        assert len(sig_inspect_spy) == 1

    def create_test_run_monitor(self, process_manager, add_data=True):
        from badger.gui.default.components.run_monitor import BadgerOptMonitor
        from badger.tests.utils import create_routine

        monitor = BadgerOptMonitor(process_manager)
        monitor.testing = True

        routine = create_routine()
        if add_data:
            routine.random_evaluate(10)
        monitor.init_plots(routine)

        if add_data:
            assert len(routine.data) == 10

        return monitor

    def test_x_axis_specification(self, qtbot, monitor, mocker):
        # check iteration/time drop down menu
        self.add_data(monitor)

        # set inspector line index 1
        monitor.inspector_variable.setValue(1)

        # Iteration selected
        monitor.cb_plot_x.setCurrentIndex(0)

        # Test label setting
        plot_var_axis = monitor.plot_var.getAxis("bottom")
        assert plot_var_axis.label.toPlainText().strip() == "iterations"

        plot_obj_axis = monitor.plot_obj.getAxis("bottom")
        assert plot_obj_axis.label.toPlainText().strip() == "iterations"

        if monitor.vocs.constraint_names:
            plot_con_axis = monitor.plot_con.getAxis("bottom")
            assert plot_con_axis.label.toPlainText().strip() == "iterations"

        assert isinstance(monitor.inspector_objective.value(), int)
        assert isinstance(monitor.inspector_variable.value(), int)
        if monitor.vocs.constraint_names:
            assert isinstance(monitor.inspector_constraint.value(), int)

        # Time selected
        monitor.cb_plot_x.setCurrentIndex(1)

        # Test label setting
        plot_var_axis_time = monitor.plot_var.getAxis("bottom")
        assert plot_var_axis_time.label.toPlainText().strip() == "time (s)"

        plot_obj_axis_time = monitor.plot_obj.getAxis("bottom")
        assert plot_obj_axis_time.label.toPlainText().strip() == "time (s)"

        if monitor.vocs.constraint_names:
            plot_con_axis_time = monitor.plot_con.getAxis("bottom")
            assert plot_con_axis_time.label.toPlainText().strip() == "time (s)"

        mock_event = mocker.MagicMock(spec=QMouseEvent)
        mock_event._scenePos = QPointF(350, 240)

        monitor.on_mouse_click(mock_event)

        # Check type of value
        assert isinstance(monitor.inspector_objective.value(), float)
        assert isinstance(monitor.inspector_variable.value(), float)
        if monitor.vocs.constraint_names:
            assert isinstance(monitor.inspector_constraint.value(), float)

        # Switch between time and iterations and see if index changes
        current_index = monitor.inspector_variable.value()

        monitor.cb_plot_x.setCurrentIndex(0)
        assert current_index != monitor.inspector_variable.value()

        monitor.cb_plot_x.setCurrentIndex(1)
        assert current_index == monitor.inspector_variable.value()

    def test_y_axis_specification(self, qtbot, monitor):
        monitor.termination_condition = {
            "tc_idx": 0,
            "max_eval": 10,
        }
        monitor.start(True)

        # Wait until the run is done
        while monitor.running:
            qtbot.wait(100)

        select_x_plot_y_axis_spy = QSignalSpy(monitor.cb_plot_y.currentIndexChanged)
        index = monitor.inspector_variable.value()

        monitor.check_relative.setChecked(False)

        # check raw - non relative
        monitor.cb_plot_y.setCurrentIndex(0)
        assert len(select_x_plot_y_axis_spy) == 0  # since 0 is the default value
        raw_value = monitor.curves_variable["x0"].getData()[1][index]
        assert raw_value == 0.5

        # relative
        monitor.check_relative.setChecked(True)

        # check non normalized relative.
        relative_value = monitor.curves_variable["x0"].getData()[1][index]
        assert relative_value == 0.0

        # normalized relative
        monitor.cb_plot_y.setCurrentIndex(1)
        assert len(select_x_plot_y_axis_spy) == 1

        normalized_relative_value = monitor.curves_variable["x0"].getData()[1][index]
        assert normalized_relative_value == 0.0

        # raw normalized
        monitor.check_relative.setChecked(False)

        normalized_raw_value = monitor.curves_variable["x0"].getData()[1][index]
        assert normalized_raw_value == 0.75

    def test_pause_play(self, qtbot, home_page):
        monitor = home_page.run_monitor
        action_bar = home_page.run_action_bar

        monitor.termination_condition = {
            "tc_idx": 0,
            "max_eval": 10,
        }
        spy = QSignalSpy(monitor.sig_pause)

        monitor.start(True)
        # qtbot.wait(500)

        qtbot.mouseClick(action_bar.btn_ctrl, Qt.MouseButton.LeftButton)
        assert len(spy) == 1

        qtbot.wait(500)

        qtbot.mouseClick(action_bar.btn_ctrl, Qt.MouseButton.LeftButton)
        assert len(spy) == 2

        while monitor.running:
            qtbot.wait(100)

    def test_jump_to_optimum(self, qtbot, home_page):
        monitor = home_page.run_monitor
        action_bar = home_page.run_action_bar

        self.add_data(monitor)
        spy = QSignalSpy(action_bar.btn_opt.clicked)
        qtbot.mouseClick(action_bar.btn_opt, Qt.MouseButton.LeftButton)

        qtbot.wait(500)

        data = monitor.routine.sorted_data

        max_value = data["f"].max()
        optimal_value_idx = monitor.inspector_variable.value()
        optimal_value = monitor.routine.sorted_data["f"][optimal_value_idx]

        # Check if signal is triggered
        assert len(spy) == 1

        # Check if it is going to be the optimal solution
        assert max_value == optimal_value

    def test_reset_environment(self, qtbot, init_multiprocessing):
        from badger.archive import save_tmp_run
        from badger.gui.acr.windows.main_window import BadgerMainWindow
        from badger.tests.utils import (
            create_routine,
            get_current_vars,
            get_vars_in_row,
        )

        window = BadgerMainWindow()

        loop = QEventLoop()
        QTimer.singleShot(1000, loop.quit)  # 1000 ms pause
        loop.exec_()

        # Run a routine
        routine = create_routine()
        tmp_filename = save_tmp_run(routine)
        home_page = window.home_page
        home_page.current_routine = routine
        monitor = home_page.run_monitor
        monitor.routine_filename = tmp_filename
        monitor.testing = True
        action_bar = home_page.run_action_bar

        # check if reset button click signal is trigged and if state is same as original state after click
        init_vars = get_current_vars(routine)

        monitor.termination_condition = {
            "tc_idx": 0,
            "max_eval": 10,
        }
        home_page.go_run(-1)
        monitor.start(True)

        while monitor.running:
            qtbot.wait(100)

        assert len(monitor.routine.data) == 10

        monitor.inspector_objective.setValue(9)

        with patch(
            "PyQt5.QtWidgets.QMessageBox.question", return_value=QMessageBox.Yes
        ):
            # with patch("PyQt5.QtWidgets.QMessageBox.information") as mock_info:
            qtbot.mouseClick(action_bar.btn_set, Qt.MouseButton.LeftButton)
            # mock_info.assert_called_once()

        # Check if current env vars matches the last solution in data
        last_vars = get_vars_in_row(monitor.routine, idx=-1)
        curr_vars = get_current_vars(monitor.routine)

        assert np.all(curr_vars == last_vars)

        # Reset env and confirm
        spy = QSignalSpy(action_bar.btn_reset.clicked)

        with patch(
            "PyQt5.QtWidgets.QMessageBox.question", return_value=QMessageBox.Yes
        ):
            # with patch("PyQt5.QtWidgets.QMessageBox.information") as mock_info:
            qtbot.mouseClick(action_bar.btn_reset, Qt.MouseButton.LeftButton)
            # mock_info.assert_called_once()

        assert len(spy) == 1

        # Check if the env has been reset
        curr_vars = get_current_vars(monitor.routine)
        assert np.all(curr_vars == init_vars)
        window.process_manager.close_proccesses()

    def test_dial_in_solution(self, qtbot, home_page):
        monitor = home_page.run_monitor
        action_bar = home_page.run_action_bar

        from badger.tests.utils import get_current_vars, get_vars_in_row

        self.add_data(monitor)

        # Check if current env vars matches the last solution in data
        last_vars = get_vars_in_row(monitor.routine, idx=-1)
        curr_vars = get_current_vars(monitor.routine)
        assert np.all(curr_vars == last_vars)

        # Dial in the third solution
        current_x_view_range = monitor.plot_var.getViewBox().viewRange()[0]

        monitor.inspector_objective.setValue(2)

        spy = QSignalSpy(action_bar.btn_set.clicked)
        with patch(
            "PyQt5.QtWidgets.QMessageBox.question", return_value=QMessageBox.Yes
        ):
            qtbot.mouseClick(action_bar.btn_set, Qt.MouseButton.LeftButton)
        assert len(spy) == 1

        new_x_view_range = monitor.plot_var.getViewBox().viewRange()[0]

        assert new_x_view_range != current_x_view_range

        # Test if the solution has been dialed in
        third_vars = get_vars_in_row(monitor.routine, idx=2)
        curr_vars = get_current_vars(monitor.routine)
        assert np.all(curr_vars == third_vars)

        # monitor.plot_x_axis = False

        # with patch("PyQt5.QtWidgets.QMessageBox.question", return_value=QMessageBox.Yes):
        #     qtbot.mouseClick(monitor.btn_set, Qt.MouseButton.LeftButton)

        # not_time_x_view_range = monitor.plot_var.getViewBox().viewRange()[0]

        # assert new_x_view_range != not_time_x_view_range

    def test_run_until(self, qtbot, home_page):
        # TODO: get this test to pass
        return
        monitor = home_page.run_monitor
        action_bar = home_page.run_action_bar

        def handle_dialog():
            while monitor.tc_dialog is None:
                QApplication.processEvents()

            # Set max evaluation to 5, then run the optimization
            monitor.tc_dialog.sb_max_eval.setValue(5)
            qtbot.mouseClick(monitor.tc_dialog.btn_run, Qt.MouseButton.LeftButton)

        QTimer.singleShot(0, handle_dialog)
        action_bar.run_until_action.trigger()

        # Wait until the run is done
        while monitor.running:
            qtbot.wait(100)

        assert len(monitor.routine.data) == 5

    """
    def test_add_extensions(self, qtbot, process_manager, init_multiprocessing):
        from badger.gui.default.components.analysis_extensions import ParetoFrontViewer
        from badger.gui.default.components.run_monitor import BadgerOptMonitor
        from badger.tests.utils import create_routine

        routine = create_routine()
        save_routine(routine)

        routine.vocs.objectives = {"f1": "MINIMIZE", "f2": "MAXIMIZE"}

        # test w/o using qtbot
        monitor = BadgerOptMonitor(process_manager)
        monitor.routine = routine

        monitor.open_extensions_palette()
        monitor.extensions_palette.add_pf_viewer()

        assert isinstance(monitor.active_extensions[0], ParetoFrontViewer)

        # TODO: logic has been changed, if ext is not applicable it won't be
        # added to the extensions palette. In order to test we need to feed in
        # a MO run here

        # test opening and closing windows
        # monitor = BadgerOptMonitor(process_manager)
        # qtbot.addWidget(monitor)

        # qtbot.mouseClick(monitor.btn_open_extensions_palette, Qt.LeftButton)
        # qtbot.mouseClick(monitor.extensions_palette.btn_data_viewer, Qt.LeftButton)
        # assert isinstance(monitor.active_extensions[0], ParetoFrontViewer)
        # assert len(monitor.active_extensions) == 1

        # test closing window -- should remove element from active extensions
        # monitor.active_extensions[0].close()
        # assert len(monitor.active_extensions) == 0
        # assert monitor.extensions_palette.n_active_extensions == 0
    """

    def test_critical_constraints(self, qtbot, home_page_critical):
        # TODO: get this test to pass
        return

        monitor = home_page_critical.run_monitor
        action_bar = home_page_critical.run_action_bar

        def handle_dialog():
            while monitor.tc_dialog is None:
                QApplication.processEvents()

            # Set max evaluation to 5, then run the optimization
            monitor.tc_dialog.sb_max_eval.setValue(5)

            qtbot.mouseClick(monitor.tc_dialog.btn_run, Qt.MouseButton.LeftButton)

        QTimer.singleShot(0, handle_dialog)
        action_bar.run_until_action.trigger()

        # Check if critical violation alert being triggered
        with patch("PyQt5.QtWidgets.QMessageBox.warning", return_value=QMessageBox.Yes):
            while monitor.running:
                qtbot.wait(100)

        assert len(monitor.routine.data) == 1  # early-termination due to violation

    def test_ucb_user_warning(self, init_multiprocessing):
        from badger.tests.utils import create_routine_constrained_ucb

        with warnings.catch_warnings(record=True) as caught_warnings:
            _ = create_routine_constrained_ucb()

            # Check if the user warning is caught
            assert len(caught_warnings) == 1
            assert isinstance(caught_warnings[0].message, UserWarning)


# TODO: Test if logbook entry is created correctly and put into the
# correct location when the logbook button is clicked
def test_send_to_logbook(qtbot):
    pass


# TODO: Test if the overlay is shown/hiden correctly
# when start/stop run button is clicked
def test_run_stop_overlay(qtbot):
    pass



from typing import Dict, Optional

from PyQt5.QtCore import pyqtSignal, QObject


class ProcessManager(QObject):
    """
    The ProcessManager class is for holding an array of live processes
    which can be used by Badger to run optimizations.
    """

    processQueueUpdated = pyqtSignal(object)

    def __init__(self) -> None:
        super().__init__()
        self.processes_queue = []

    def add_to_queue(self, process_with_args: Dict) -> None:
        """
        Add to a dict contaitng a process and it's coresponding args to the processes_queue.

        Parameters
        ----------
        process_with_args: dict
        """
        self.processes_queue.append(process_with_args)

    def remove_from_queue(self) -> Optional[Dict]:
        """
        Removes and returns a process and it's coresponding args to the processes_queue.
        If no process are in the processes_queue then the method returns None.

        Returns
        -------
        process_with_args: dict | None
        """
        if self.processes_queue:
            process_with_args = self.processes_queue.pop(0)
            self.processQueueUpdated.emit(self.processes_queue)
            return process_with_args

        return None

    def close_proccesses(self) -> bool:
        """
        Closes the processes stored in the processes_queue.

        Returns
        -------
        True: bool
        """
        for i in range(0, len(self.processes_queue)):
            process = self.processes_queue.pop(0)
            process["process"].terminate()
            process["process"].join()

        return True



import os
import traceback
from importlib import resources
from typing import List

import numpy as np
import pandas as pd
import pyqtgraph as pg
from PyQt5.QtCore import pyqtSignal
from pyqtgraph.Qt import QtGui, QtCore

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (
    QCheckBox,
    QComboBox,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QStyledItemDelegate,
    QVBoxLayout,
    QWidget,
)
from xopt import VOCS

from badger.archive import archive_run, BADGER_ARCHIVE_ROOT

# from ...utils import AURORA_PALETTE, FROST_PALETTE
from badger.logbook import BADGER_LOGBOOK_ROOT, send_to_logbook
from badger.routine import Routine
from badger.tests.utils import get_current_vars
from badger.gui.default.windows.message_dialog import BadgerScrollableMessageBox
from badger.gui.default.windows.terminition_condition_dialog import (
    BadgerTerminationConditionDialog,
)

from badger.gui.default.components.extensions_palette import ExtensionsPalette
from badger.gui.default.components.routine_runner import BadgerRoutineSubprocess

import logging

logger = logging.getLogger(__name__)

# disable chained assignment warning from pydantic
pd.options.mode.chained_assignment = None  # default='warn'


class BadgerOptMonitor(QWidget):
    sig_pause = pyqtSignal(bool)  # True: pause, False: resume
    sig_stop = pyqtSignal()
    sig_lock = pyqtSignal(bool)  # True: lock GUI, False: unlock GUI
    sig_new_run = pyqtSignal()  # start the new run
    sig_run_started = pyqtSignal()  # run started
    sig_stop_run = pyqtSignal()  # stop the run
    sig_run_name = pyqtSignal(str)  # filename of the new run
    sig_status = pyqtSignal(str)  # status information
    sig_inspect = pyqtSignal(int)  # index of the inspector
    sig_progress = pyqtSignal(pd.DataFrame)  # new evaluated solution
    sig_del = pyqtSignal()
    sig_routine_finished = pyqtSignal()
    sig_lock_action = pyqtSignal()
    sig_toggle_reset = pyqtSignal(bool)
    sig_toggle_run = pyqtSignal(bool)
    sig_toggle_other = pyqtSignal(bool)
    sig_env_ready = pyqtSignal()

    def __init__(self, process_manager=None):
        super().__init__()
        # self.setAttribute(Qt.WA_DeleteOnClose, True)

        # For plot type switching
        self.x_plot_y_axis = 0  # 0: raw, 1: normalized
        self.plot_x_axis = 0  # 0: iteration, 1: time
        self.x_plot_relative = True
        # Routine info
        self.routine = None
        self.routine_filename = None
        self.process_manager = process_manager

        # Curves in the monitor
        self.curves_variable = {}
        self.curves_objective = {}
        self.curves_constraint = {}
        self.curves_sta = {}

        # Run optimization
        self.routine_runner = None
        self.running = False
        # Fix the auto range issue
        self.eval_count = 0
        # Termination condition for the run
        self.termination_condition = None

        self.extensions_palette = ExtensionsPalette(self)
        self.active_extensions = []

        self.testing = False
        self.tc_dialog = None
        self.post_run_actions = []

        self.init_ui()
        self.config_logic()
        self._states = None

    @property
    def vocs(self) -> VOCS:
        return self.routine.vocs

    def states(self, new_states: dict) -> None:
        self._states = new_states

    def init_ui(self):
        # Load all icons
        icon_ref = resources.files(__package__) / "../images/play.png"
        with resources.as_file(icon_ref) as icon_path:
            self.icon_play = QIcon(str(icon_path))
        icon_ref = resources.files(__package__) / "../images/pause.png"
        with resources.as_file(icon_ref) as icon_path:
            self.icon_pause = QIcon(str(icon_path))
        icon_ref = resources.files(__package__) / "../images/stop.png"
        with resources.as_file(icon_ref) as icon_path:
            self.icon_stop = QIcon(str(icon_path))

        # self.main_panel = main_panel = QWidget(self)
        # main_panel.setStyleSheet('background-color: #19232D;')
        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(0, 8, 0, 0)

        # Config bar
        config_bar = QWidget()
        hbox_config = QHBoxLayout(config_bar)
        hbox_config.setContentsMargins(0, 0, 0, 0)
        label = QLabel("Evaluation History Plot Type")
        label_x = QLabel("X Axis")
        self.cb_plot_x = cb_plot_x = QComboBox()
        cb_plot_x.setItemDelegate(QStyledItemDelegate())
        cb_plot_x.addItems(["Iteration", "Time"])
        label_y = QLabel("Y Axis (Var)")
        self.cb_plot_y = cb_plot_y = QComboBox()
        cb_plot_y.setItemDelegate(QStyledItemDelegate())
        cb_plot_y.addItems(["Raw", "Normalized"])
        self.check_relative = check_relative = QCheckBox("Relative")
        check_relative.setChecked(True)
        hbox_config.addWidget(label)
        # hbox_config.addSpacing(1)
        hbox_config.addWidget(label_x)
        hbox_config.addWidget(cb_plot_x, 1)
        hbox_config.addWidget(label_y)
        hbox_config.addWidget(cb_plot_y, 1)
        hbox_config.addWidget(check_relative)

        # Set up the monitor
        # Don't set show=True or there will be a blank window flashing once
        self.monitor = monitor = pg.GraphicsLayoutWidget()
        pg.setConfigOptions(antialias=True)
        # monitor.ci.setBorder((50, 50, 100))
        # monitor.resize(1000, 600)

        # create vertical cursor lines
        self.inspector_objective = create_cursor_line()
        self.inspector_constraint = create_cursor_line()
        self.inspector_state = create_cursor_line()
        self.inspector_variable = create_cursor_line()

        # add axes
        self.plot_obj = plot_obj = add_axes(
            monitor, "objectives", "Evaluation History (Y)", self.inspector_objective
        )

        monitor.nextRow()  # leave space for the cons plot
        monitor.nextRow()  # leave space for the stas plot
        monitor.nextRow()

        self.plot_var = plot_var = add_axes(
            monitor,
            "Relative Variable Value",
            "Relative Variable History (X)",
            self.inspector_variable,
        )

        plot_var.setXLink(plot_obj)

        self.colors = ["c", "g", "m", "y", "b", "r", "w"]
        self.symbols = ["o", "t", "t1", "s", "p", "h", "d"]

        vbox.addWidget(config_bar)
        vbox.addWidget(monitor)

    # noinspection PyUnresolvedReferences
    def config_logic(self):
        """
        Configure the logic and connections for various interactive elements in the
        application.

        This method sets up event connections and handlers for different interactive
        elements in the application, such as the inspector lines, buttons,
        and visualizations. It establishes connections between signals and slots to
        enable user interaction and control of the application's functionality.

        Notes
        -----
        - The `config_logic` method is intended to be called once during
        the initialization of the application or a specific class. It sets up various
        event handlers and connections for interactive elements.

        - The signals and slots established in this method determine how the
        application responds to user actions, such as button clicks, inspector line
        drags, and selection changes in visualization controls.

        - Ensure that the necessary attributes and dependencies are properly
        initialized before calling this method.

        """

        # Sync the inspector lines
        self.inspector_objective.sigDragged.connect(self.ins_obj_dragged)
        self.inspector_objective.sigPositionChangeFinished.connect(self.ins_drag_done)
        self.inspector_constraint.sigDragged.connect(self.ins_con_dragged)
        self.inspector_constraint.sigPositionChangeFinished.connect(self.ins_drag_done)
        self.inspector_state.sigDragged.connect(self.ins_sta_dragged)
        self.inspector_state.sigPositionChangeFinished.connect(self.ins_drag_done)
        self.inspector_variable.sigDragged.connect(self.ins_var_dragged)
        self.inspector_variable.sigPositionChangeFinished.connect(self.ins_drag_done)
        self.plot_obj.scene().sigMouseClicked.connect(self.on_mouse_click)
        # sigMouseReleased.connect(self.on_mouse_click)

        # Visualization
        self.cb_plot_x.currentIndexChanged.connect(self.select_x_axis)
        self.cb_plot_y.currentIndexChanged.connect(self.select_x_plot_y_axis)
        self.check_relative.stateChanged.connect(self.toggle_x_plot_y_axis_relative)

    def init_plots(self, routine: Routine = None, run_filename: str = None):
        """
        Initialize and configure the plots and related components in the application.

        This method initializes and configures the plot areas, curves, and inspectors
        used for visualizing data in the application. It also manages the state of
        various UI elements based on the provided routine and run information.

        Parameters
        ----------
        routine : Routine,
            The routine to use for configuring the plots. If
            not provided, the method will use the previously set routine.

        run_filename : str, optional
            The filename of the run, used to determine the state of the application's UI
            elements.

        Returns
        -------
        None

        Notes
        -----
        - The `init_plots` method is typically called during the application's
        initialization or when a new routine is selected. It sets up
        the plots, curves, and inspectors for visualizing data.

        - The method relies on the `self.routine`, `self.vocs`, `self.monitor`,
        `self.colors`, `self.symbols`, and other attributes of the class to configure
        the plots and manage UI elements.

        - The `routine` parameter allows you to provide a specific routine for
        configuring the plots. If not provided, it will use the previously set routine.

        - The `run_filename` parameter is used to determine the state of UI elements,
        such as enabling or disabling certain buttons based on whether the run data
        is available.

        """
        if routine is None:
            # if no routines are specified, clear the current plots
            self.plot_var.clear()
            self.plot_obj.clear()

            # if constraints are active delete them
            try:
                self.monitor.removeItem(self.plot_con)
                self.plot_con.removeItem(self.inspector_constraint)
                del self.plot_con
            except:
                pass

            # if statics exist delete that plot
            try:
                self.monitor.removeItem(self.plot_obs)
                self.plot_obs.removeItem(self.inspector_state)
                del self.plot_obs
            except:
                pass

            # if no routine is loaded set button to disabled
            self.sig_lock_action.emit()

            self.routine = None

            return

        self.routine = routine

        # Retrieve data information
        objective_names = self.vocs.objective_names
        variable_names = self.vocs.variable_names
        constraint_names = self.vocs.constraint_names
        sta_names = self.vocs.observable_names

        # Configure variable plots
        self.curves_variable = self._configure_plot(
            self.plot_var, self.inspector_variable, variable_names
        )

        # Configure objective plots
        self.curves_objective = self._configure_plot(
            self.plot_obj, self.inspector_objective, objective_names
        )

        # Configure constraint plots
        if constraint_names:
            try:
                self.plot_con
            except:
                self.plot_con = plot_con = add_axes(
                    self.monitor,
                    "constraints",
                    "Evaluation History (C)",
                    self.inspector_constraint,
                    row=1,
                    col=0,
                )
                plot_con.setXLink(self.plot_obj)

            self.curves_constraint = self._configure_plot(
                self.plot_con, self.inspector_constraint, constraint_names
            )

        else:
            try:
                self.monitor.removeItem(self.plot_con)
                self.plot_con.removeItem(self.inspector_constraint)
                del self.plot_con
            except:
                pass

        # Configure state plots
        if sta_names:
            try:
                self.plot_obs
            except:
                self.plot_obs = plot_obs = add_axes(
                    self.monitor,
                    "observables",
                    "Evaluation History (S)",
                    self.inspector_state,
                    row=2,
                    col=0,
                )
                plot_obs.setXLink(self.plot_obj)

            self.curves_sta = self._configure_plot(
                self.plot_obs, self.inspector_state, sta_names
            )
        else:
            try:
                self.monitor.removeItem(self.plot_obs)
                self.plot_obs.removeItem(self.inspector_state)
                del self.plot_obs
            except:
                pass

        # Reset inspectors
        self.inspector_objective.setValue(0)
        self.inspector_variable.setValue(0)
        self.inspector_constraint.setValue(0)
        self.inspector_state.setValue(0)

        # Switch run button state
        self.sig_toggle_run.emit(False)

        self.eval_count = 0  # reset the evaluation count
        self.enable_auto_range()

        # Reset button should only be available if it's the current run
        if self.routine_runner and self.routine_runner.run_filename == run_filename:
            self.sig_toggle_reset.emit(False)
        else:
            self.sig_toggle_reset.emit(True)

        if routine.data is None:
            self.sig_toggle_other.emit(True)

            return

        self.update_curves()

        self.sig_toggle_other.emit(False)

    def _configure_plot(self, plot_object, inspector, names):
        plot_object.clear()
        plot_object.addItem(inspector)
        curves = {}
        for i, name in enumerate(names):
            color = self.colors[i % len(self.colors)]
            # symbol = self.symbols[i % len(self.colors)]

            # add a dot symbol to the plot to handle cases were there are many nans
            dot_symbol = QtGui.QPainterPath()
            size = 0.075  # size of the dot symbol
            dot_symbol.addEllipse(QtCore.QRectF(-size / 2, -size / 2, size, size))

            pen = pg.mkPen(color, width=3)
            _curve = plot_object.plot(
                pen=pen,
                symbol=dot_symbol,
                symbolPen=pen,
                name=name,
                symbolBrush=pen.color(),
            )
            curves[name] = _curve

        return curves

    def init_routine_runner(self):
        self.reset_routine_runner()

        self.routine_runner = routine_runner = BadgerRoutineSubprocess(
            self.process_manager,
            self.routine,
            self.routine_filename,
            save=True,
            testing=self.testing,
        )

        routine_runner.signals.env_ready.connect(self.env_ready)
        routine_runner.signals.finished.connect(self.routine_finished)
        routine_runner.signals.progress.connect(self.update)
        routine_runner.signals.error.connect(self.on_error)
        routine_runner.signals.info.connect(self.on_info)
        routine_runner.signals.states.connect(self.states)

        self.sig_pause.connect(routine_runner.ctrl_routine)
        self.sig_stop.connect(routine_runner.stop_routine)

    def reset_routine_runner(self):
        if self.routine_runner:
            self.sig_pause.disconnect()
            self.sig_stop.disconnect()
            self.routine_runner = None

    def start(self, use_termination_condition=False):
        self.sig_new_run.emit()
        self.sig_status.emit(f"Running routine {self.routine.name}...")
        self.routine.data = None  # reset data if any
        self.init_plots(self.routine)
        self.init_routine_runner()
        if use_termination_condition:
            self.routine_runner.set_termination_condition(self.termination_condition)
        self.running = True  # if a routine runner is working
        self.routine_runner.run()
        self.sig_run_started.emit()
        self.sig_lock.emit(True)

    def save_termination_condition(self, tc):
        self.termination_condition = tc

    def enable_auto_range(self):
        # Enable autorange
        self.plot_obj.enableAutoRange()
        self.plot_var.enableAutoRange()
        if self.vocs.constraint_names:
            self.plot_con.enableAutoRange()

        if self.vocs.observable_names:
            self.plot_obs.enableAutoRange()

    def open_extensions_palette(self):
        self.extensions_palette.show()

    def extension_window_closed(self, child_window):
        self.active_extensions.remove(child_window)
        self.extensions_palette.update_palette()

    def extract_timestamp(self, data=None):
        if data is None:
            data = self.routine.sorted_data

        return data["timestamp"].to_numpy(copy=True)

    def update(self, results: pd.DataFrame) -> None:
        """Update plots in main window as well as any active extensions and the
        extensions palette

        Parameters
        ----------
        results : pd.DataFrame

        Returns
        -------
        None
        """
        self.update_curves(results)
        self.update_analysis_extensions()
        self.extensions_palette.update_palette()

        # Quick-n-dirty fix to the auto range issue
        self.eval_count += 1
        if self.eval_count < 2:
            self.enable_auto_range()

        self.sig_progress.emit(self.routine.data.tail(1))

        # Check critical condition
        self.check_critical()

    def update_curves(self, results=None):
        use_time_axis = self.plot_x_axis == 1
        normalize_inputs = self.x_plot_y_axis == 1

        if results is not None:
            self.routine.data = results

        if not self.routine or not hasattr(self.routine, "sorted_data"):
            # if no routine or sorted_data is available, return
            return
        data_copy = self.routine.sorted_data

        # Get timestamps
        if use_time_axis:
            ts = self.extract_timestamp(data_copy)
            ts -= ts[0]
        else:
            ts = None

        variable_names = self.vocs.variable_names

        # if normalize x, normalize using vocs
        if normalize_inputs:
            input_data = self.vocs.normalize_inputs(data_copy)
        else:
            input_data = data_copy[variable_names]

        # if plot relative, subtract the first value from the dict
        if self.x_plot_relative:
            input_data[variable_names] = (
                input_data[variable_names] - input_data[variable_names].iloc[0]
            )

        set_data(variable_names, self.curves_variable, input_data, ts)
        set_data(self.vocs.objective_names, self.curves_objective, data_copy, ts)
        set_data(self.vocs.constraint_names, self.curves_constraint, data_copy, ts)
        set_data(self.vocs.observable_names, self.curves_sta, data_copy, ts)

    def check_critical(self) -> None:
        """
        Check if a critical constraint has been violated in the last data point,
        and take appropriate actions if so.

        If there are no critical constraints, the function will return without taking
        any action. If a critical constraint has been violated, it will pause the
        run, open a dialog to inform the user about the violation, and provide
        options to terminate or resume the run.

        Returns
        -------
        None

        Notes
        -----

        The critical constraints are determined by the
        `self.routine.critical_constraint_names` attribute. If no critical
        constraints are defined, this function will have no effect.

        The function emits signals `self.sig_pause` and `self.sig_stop` to handle the
        pause and stop actions.

        """
        # if there are no critical constraints then skip
        if len(self.routine.critical_constraint_names) == 0:
            return

        feas = self.vocs.feasibility_data(self.routine.data.tail(1), prefix="")
        feasible = feas["feasible"].iloc[0].item()
        if feasible:
            return

        # if code reaches this point there is a critical constraint violated
        self.sig_pause.emit(True)
        # self.btn_ctrl.setIcon(self.icon_play)
        # self.btn_ctrl.setToolTip("Resume")
        # self.btn_ctrl._status = "play"

        # Show the list of critical violated constraints
        feas_crit = feas[self.routine.critical_constraint_names]
        violated_crit = feas_crit.columns[~feas_crit.iloc[0]].tolist()
        msg = "\n".join(violated_crit)
        reply = QMessageBox.warning(
            self,
            "Run Paused",
            f"The following critical constraints were violated:\n\n{msg}\n\nTerminate the run?",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.Yes,
        )
        if reply == QMessageBox.Yes:
            self.sig_stop.emit()
            self.sig_stop_run.emit()

    def update_analysis_extensions(self) -> None:
        for ele in self.active_extensions:
            try:
                ele.update_window(self.routine)
            except ValueError:
                traceback.print_exc()

    def env_ready(self, init_vars) -> None:
        self.init_vars = init_vars

        self.sig_env_ready.emit()

    def routine_finished(self) -> None:
        self.running = False
        self.sig_routine_finished.emit()

        self.sig_lock.emit(False)

        try:
            # TODO: fill in the states
            # TODO: replace self.testing with with another processes for having a testing mode
            if not self.testing:
                run = archive_run(self.routine, states=self._states)
                self.routine_runner.run_filename = run["filename"]
                env = self.routine.environment
                path = run["path"]
                filename = run["filename"][:-4] + "pickle"

            try:
                env.interface.stop_recording(os.path.join(path, filename))
            except AttributeError:  # recording was not enabled
                pass

            self.sig_run_name.emit(run["filename"])
            self.sig_status.emit(
                f"Archive success: Run data archived to {BADGER_ARCHIVE_ROOT}"
            )
            # if not self.testing:
            #     QMessageBox.information(
            #         self, 'Success!',
            #         f'Archive success: Run data archived to {BADGER_ARCHIVE_ROOT}')

        except Exception as e:
            self.sig_run_name.emit(None)
            self.sig_status.emit(f"Archive failed: {str(e)}")
            # if not self.testing:
            #     QMessageBox.critical(self, 'Archive failed!',
            #                          f'Archive failed: {str(e)}')
        finally:
            for action in self.post_run_actions:
                action()

        # self.reset_routine_runner()

    def destroy_unused_env(self) -> None:
        if not self.running:
            try:
                del self.routine_runner.routine.environment
            except AttributeError:  # env already destroyed
                pass

            try:
                del self.routine.environment
            except AttributeError:  # env already destroyed
                pass

    def on_error(self, error):
        details = error._details if hasattr(error, "_details") else None

        dialog = BadgerScrollableMessageBox(
            title="Error!", text=str(error), parent=self
        )
        dialog.setIcon(QMessageBox.Critical)
        dialog.setDetailedText(details)
        dialog.exec_()

    # Do not show info -- too distracting
    def on_info(self, msg):
        pass

    def logbook(self):
        try:
            send_to_logbook(self.routine, self.monitor)
        except Exception as e:
            self.sig_status.emit(f"Log failed: {str(e)}")
            # QMessageBox.critical(self, 'Log failed!', str(e))

            return

        self.sig_status.emit(f"Log success: Log saved to {BADGER_LOGBOOK_ROOT}")
        # QMessageBox.information(
        #     self, 'Success!', f'')

    def ctrl_routine(self, status):
        self.sig_pause.emit(status)

    def ins_obj_dragged(self, ins_obj):
        self.inspector_variable.setValue(ins_obj.value())
        if self.vocs.constraint_names:
            self.inspector_constraint.setValue(ins_obj.value())
        if self.vocs.observable_names:
            self.inspector_state.setValue(ins_obj.value())

    def ins_con_dragged(self, ins_con):
        self.inspector_variable.setValue(ins_con.value())
        self.inspector_objective.setValue(ins_con.value())
        if self.vocs.observable_names:
            self.inspector_state.setValue(ins_con.value())

    def ins_sta_dragged(self, ins_sta):
        self.inspector_variable.setValue(ins_sta.value())
        self.inspector_objective.setValue(ins_sta.value())
        if self.vocs.constraint_names:
            self.inspector_constraint.setValue(ins_sta.value())

    def ins_var_dragged(self, ins_var):
        self.inspector_objective.setValue(ins_var.value())
        if self.vocs.constraint_names:
            self.inspector_constraint.setValue(ins_var.value())
        if self.vocs.observable_names:
            self.inspector_state.setValue(ins_var.value())

    def ins_drag_done(self, ins):
        self.sync_ins(ins.value())

    def sync_ins(self, pos):
        if self.plot_x_axis:  # x-axis is time
            value, idx = self.closest_ts(pos)
        else:
            try:
                ts = self.extract_timestamp()
                value = idx = np.clip(np.round(pos), 0, len(ts) - 1)
            except:  # no data
                value = idx = np.round(pos)
        self.inspector_objective.setValue(value)
        if self.vocs and self.vocs.constraint_names:
            self.inspector_constraint.setValue(value)
        if self.vocs and self.vocs.observable_names:
            self.inspector_state.setValue(value)
        self.inspector_variable.setValue(value)

        self.sig_inspect.emit(int(idx))

    def closest_ts(self, t):
        # Get the closest timestamp in data regarding t
        ts = self.extract_timestamp()
        ts -= ts[0]
        idx = np.argmin(np.abs(ts - t))

        return ts[idx], idx

    def reset_env(self):
        reply = QMessageBox.question(
            self,
            "Reset Environment",
            f"Are you sure you want to reset the env vars back to {self.init_vars}?",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )
        if reply != QMessageBox.Yes:
            return

        curr_vars = get_current_vars(self.routine)

        self.routine.environment._set_variables(
            dict(zip(self.vocs.variable_names, self.init_vars))
        )

        self.jump_to_solution(0)
        self.sig_inspect.emit(0)
        # Center around the zero position
        if self.plot_x_axis:  # x-axis is time
            pos, _ = self.closest_ts(self.inspector_objective.value())
        else:
            pos = int(self.inspector_objective.value())
        x_range = self.plot_var.getViewBox().viewRange()[0]
        delta = (x_range[1] - x_range[0]) / 2
        self.plot_var.setXRange(pos - delta, pos + delta, padding=0)

        self.sig_status.emit(
            f"Reset environment: Env vars {curr_vars} -> {self.init_vars}"
        )
        # QMessageBox.information(self, 'Reset Environment',
        #                         f'Env vars {curr_vars} -> {self.init_vars}')

    def jump_to_optimal(self):
        try:
            best_idx, _, _ = self.routine.vocs.select_best(
                self.routine.sorted_data, n=1
            )
            # print(best_idx, _)
            best_idx = int(best_idx[0])

            self.jump_to_solution(best_idx)
            self.sig_inspect.emit(best_idx)
        except NotImplementedError:
            QMessageBox.warning(
                self,
                "Jump to optimum",
                "Jump to optimum is not supported for multi-objective optimization yet",
            )

    def jump_to_solution(self, idx):
        if self.plot_x_axis:  # x-axis is time
            ts = self.extract_timestamp()
            value = ts[idx] - ts[0]
        else:
            value = idx

        self.inspector_objective.setValue(value)
        if self.vocs.constraint_names:
            self.inspector_constraint.setValue(value)
        if self.vocs.observable_names:
            self.inspector_state.setValue(value)
        self.inspector_variable.setValue(value)

    def set_vars(self):
        df = self.routine.sorted_data
        if self.plot_x_axis:  # x-axis is time
            pos, idx = self.closest_ts(self.inspector_objective.value())
        else:
            pos = idx = int(self.inspector_objective.value())
        variable_names = self.vocs.variable_names
        solution = df[variable_names].to_numpy()[idx]
        curr_vars = get_current_vars(self.routine)

        reply = QMessageBox.question(
            self,
            "Apply Solution",
            "Are you sure you want to apply the selected solution:\n"
            + "\n".join(
                f"{variable_names[i]}: {round(curr_vars[i], 3)} -> {round(solution[i], 3)},"
                for i in range(len(variable_names))
            )
            + "\nto "
            + f"{self.routine.environment.name}?",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )
        if reply != QMessageBox.Yes:
            return

        curr_vars = get_current_vars(self.routine)
        # Patch the hard limit to the environment
        # TODO: this could lead to unexpected behavior
        # since we patched the class variable directly
        # there ought to be a better way to do this
        if self.routine.vrange_hard_limit:
            self.routine.environment.variables.update(self.routine.vrange_hard_limit)
        self.routine.environment._set_variables(dict(zip(variable_names, solution)))
        # center around the inspector
        x_range = self.plot_var.getViewBox().viewRange()[0]
        delta = (x_range[1] - x_range[0]) / 2
        self.plot_var.setXRange(pos - delta, pos + delta, padding=0)

        updated_vars = get_current_vars(self.routine)
        self.sig_status.emit(
            f"Dial in solution: {[f'{variable_names[i]}: {round(curr_vars[i], 4)} -> {round(updated_vars[i], 4)}' for i in range(len(variable_names))]}"
        )
        # QMessageBox.information(
        #     self, 'Set Environment', f'Env vars have been set to {solution}')

    def select_x_axis(self, i):
        self.plot_x_axis = i

        # Switch the x-axis labels
        if i:
            self.plot_var.setLabel("bottom", "time (s)")
            self.plot_obj.setLabel("bottom", "time (s)")
            if self.vocs.constraint_names:
                self.plot_con.setLabel("bottom", "time (s)")
            if self.vocs.observable_names:
                self.plot_obs.setLabel("bottom", "time (s)")
        else:
            self.plot_var.setLabel("bottom", "iterations")
            self.plot_obj.setLabel("bottom", "iterations")
            if self.vocs.constraint_names:
                self.plot_con.setLabel("bottom", "iterations")
            if self.vocs.observable_names:
                self.plot_obs.setLabel("bottom", "iterations")

        # Update inspector line position
        if i:
            ts = self.extract_timestamp()
            value = ts[int(self.inspector_objective.value())] - ts[0]
        else:
            _, value = self.closest_ts(self.inspector_objective.value())
        self.inspector_objective.setValue(value)
        if self.vocs.constraint_names:
            self.inspector_constraint.setValue(value)
        if self.vocs.observable_names:
            self.inspector_state.setValue(value)
        self.inspector_variable.setValue(value)

        self.update_curves()
        self.enable_auto_range()

    def select_x_plot_y_axis(self, i):
        self.x_plot_y_axis = i
        self.update_curves()

    def toggle_x_plot_y_axis_relative(self):
        self.x_plot_relative = self.check_relative.isChecked()

        # Change axes labels depending on if relative is checked
        if self.x_plot_relative:
            self.plot_var.setLabel("left", "Relative Variable Value")
            self.plot_var.setTitle("Relative Variable History (X)")
        else:
            self.plot_var.setLabel("left", "Variable Value")
            self.plot_var.setTitle("Variable History (X)")

        self.update_curves()

    def on_mouse_click(self, event):
        # https://stackoverflow.com/a/64081483
        coor_obj = self.plot_obj.vb.mapSceneToView(event._scenePos)
        if self.vocs and self.vocs.constraint_names:
            coor_con = self.plot_con.vb.mapSceneToView(event._scenePos)
        if self.vocs and self.vocs.observable_names:
            coor_sta = self.plot_obs.vb.mapSceneToView(event._scenePos)
        coor_var = self.plot_var.vb.mapSceneToView(event._scenePos)

        flag = self.plot_obj.viewRect().contains(
            coor_obj
        ) or self.plot_var.viewRect().contains(coor_var)
        if self.vocs and self.vocs.constraint_names:
            flag = flag or self.plot_con.viewRect().contains(coor_con)
        if self.vocs and self.vocs.observable_names:
            flag = flag or self.plot_obs.viewRect().contains(coor_sta)

        if flag:
            self.sync_ins(coor_obj.x())

    def delete_run(self):
        self.sig_del.emit()

    def stop(self):
        self.sig_stop.emit()
        self.sig_stop_run.emit()

    def start_until(self):
        dlg = BadgerTerminationConditionDialog(
            self,
            self.start,
            self.save_termination_condition,
            self.termination_condition,
        )
        self.tc_dialog = dlg
        try:
            dlg.exec()
        finally:
            self.tc_dialog = None

    def register_post_run_action(self, action):
        self.post_run_actions.append(action)


def add_axes(monitor, ylabel, title, cursor_line, **kwargs):
    plot_obj = monitor.addPlot(title=title, **kwargs)
    plot_obj.setLabel("left", ylabel)
    plot_obj.setLabel("bottom", "iterations")
    plot_obj.showGrid(x=True, y=True)
    leg_obj = plot_obj.addLegend()
    leg_obj.setBrush((50, 50, 100, 200))

    plot_obj.addItem(cursor_line)

    return plot_obj


def create_cursor_line():
    return pg.InfiniteLine(
        movable=True,
        angle=90,
        label=None,
        labelOpts={
            "position": 0.1,
            "color": (200, 200, 100),
            "fill": (200, 200, 200, 50),
            "movable": True,
        },
    )


def set_data(names: List[str], curves: dict, data: pd.DataFrame, ts=None):
    for name in names:
        if ts is not None:
            curves[name].setData(ts, data[name].to_numpy(dtype=np.double))
        else:
            curves[name].setData(data[name].to_numpy(dtype=np.double))



from abc import abstractmethod
from typing import Optional, cast

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QMessageBox
from PyQt5.QtGui import QCloseEvent
from badger.gui.default.components.bo_visualizer.bo_plotter import BOPlotWidget
from badger.gui.default.components.pf_viewer.pf_widget import ParetoFrontWidget
from badger.routine import Routine

import logging

from xopt.generators.bayesian.bayesian_generator import BayesianGenerator

logger = logging.getLogger(__name__)


class AnalysisExtension(QDialog):
    window_closed = pyqtSignal(object)

    def __init__(self, parent: Optional[QDialog] = None):
        super().__init__(parent=parent)

    @abstractmethod
    def update_window(self, routine: Routine):
        pass

    def closeEvent(self, a0: Optional[QCloseEvent]) -> None:
        self.window_closed.emit(self)
        super().closeEvent(a0)


class ParetoFrontViewer(AnalysisExtension):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("Pareto Front Viewer")

        self.pw_widget = ParetoFrontWidget()

        layout = QVBoxLayout()
        layout.addWidget(self.pw_widget)
        self.setLayout(layout)

    def update_window(self, routine: Routine):
        try:
            self.pw_widget.update_plot(routine)
        except:
            self.close()


class BOVisualizer(AnalysisExtension):
    df_length = float("inf")
    initialized = False
    correct_generator = False
    routine_identifier = ""
    plot_update_rate = 250

    def __init__(self):
        logger.debug("Initializing BO Visualizer Extension")

        super().__init__()
        self.setWindowTitle("BO Visualizer")

        # Initialize BOPlotWidget without a routine
        self.bo_plot_widget = BOPlotWidget()

        logger.debug("Initialized BOPlotWidget")

        bo_layout = QVBoxLayout()
        bo_layout.addWidget(self.bo_plot_widget)
        self.setLayout(bo_layout)

        logger.debug("Set layout for BOVisualizer")

    def requires_reinitialization(self):
        # Check if the extension needs to be reinitialized
        logger.debug("Checking if BO Visualizer needs to be reinitialized")

        if not self.initialized:
            logger.debug("Reset - Extension never initialized")
            self.initialized = True
            return True

        if self.routine_identifier != self.routine.name:
            logger.debug("Reset - Routine name has changed")
            self.identifier = self.routine.name
            return True

        if self.routine.data is None:
            logger.debug("Reset - No data available")
            return True

        previous_len = self.df_length
        self.df_length = len(self.routine.data)
        new_length = self.df_length

        if previous_len > new_length:
            logger.debug("Reset - Data length is the same or smaller")
            self.df_length = float("inf")
            return True

        return False

    def update_window(self, routine: Routine):
        # Updating the BO Visualizer window
        logger.debug("Updating BO Visualizer window")

        try:
            # Update the routine with new generator model if applicable
            self.update_routine(routine)
        except:
            self.close()

        if not self.correct_generator:
            logger.debug("Incorrect generator type")
            return

        if self.requires_reinitialization():
            self.bo_plot_widget.initialize_widget(self.routine, self.update_window)

        # Update the plots with the new generator model
        self.bo_plot_widget.update_plot(self.plot_update_rate)

    def update_routine(self, routine: Routine):
        logger.debug("Updating routine in BO Visualizer")

        self.routine = routine

        # Check if the generator is a BayesianGenerator
        if not issubclass(self.routine.generator.__class__, BayesianGenerator):
            self.correct_generator = False
            QMessageBox.critical(
                self,
                "Invalid Generator",
                f"Invalid generator type: {type(self.routine.generator)}, BO Visualizer only supports BayesianGenerator",
            )
            raise TypeError(
                f"Invalid generator type: {type(self.routine.generator)}, BO Visualizer only supports BayesianGenerator"
            )

        self.correct_generator = True

        generator = cast(BayesianGenerator, self.routine.generator)

        # Handle the edge case where the extension has been opened after an optimization has already finished.
        if generator.model is None:
            logger.warning("Model not found in generator")
            if generator.data is None:
                if self.routine.data is None:
                    logger.error("No data available in routine or generator")
                    QMessageBox.critical(
                        self,
                        "No data available",
                        "No data available in routine or generator",
                    )
                    return

                # Use the data from the routine to train the model
                logger.debug("Setting generator data from routine")
                generator.data = self.routine.data

            try:
                generator.train_model(generator.data)
            except Exception as e:
                logger.error(f"Failed to train model: {e}")
                QMessageBox.warning(
                    self,
                    "Failed to train model",
                    f"Failed to train model: {e}",
                )
        else:
            logger.debug("Model already exists in generator")

        if generator.data is None:
            logger.error("No data available in generator")
            QMessageBox.critical(
                self,
                "No data available",
                "No data available in generator",
            )
            return

        self.df_length = len(generator.data)
        self.routine.generator = generator



from PyQt5.QtWidgets import (
    QDialog,
    QWidget,
    QHBoxLayout,
    QLineEdit,
    QPushButton,
    QVBoxLayout,
)
from PyQt5.QtWidgets import QGroupBox, QMessageBox
from badger.gui.default.components.labeled_lineedit import labeled_lineedit
from badger.environment import instantiate_env


class BadgerVariableDialog(QDialog):
    def __init__(self, parent, env_class, configs, callback):
        super().__init__(parent)

        self.env = instantiate_env(env_class, configs)
        self.callback = callback

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        self.setWindowTitle("Add variable")
        self.setMinimumWidth(360)

        vbox = QVBoxLayout(self)

        # Action bar
        action_bar = QWidget()
        hbox = QHBoxLayout(action_bar)
        hbox.setContentsMargins(0, 0, 0, 0)
        self.edit_name = edit_name = QLineEdit()
        edit_name.setPlaceholderText("Variable name")
        self.btn_check = btn_check = QPushButton("Check")
        btn_check.setDisabled(True)
        btn_check.setFixedSize(96, 24)
        hbox.addWidget(edit_name, 1)
        hbox.addWidget(btn_check)

        # Info group
        group_info = QGroupBox("Variable Info")
        vbox_info = QVBoxLayout(group_info)

        self.edit_value = labeled_lineedit("value", "", 48)
        self.edit_min = labeled_lineedit("min", "", 48)
        self.edit_max = labeled_lineedit("max", "", 48)
        vbox_info.addWidget(self.edit_value)
        vbox_info.addWidget(self.edit_min)
        vbox_info.addWidget(self.edit_max)

        # Button set
        button_set = QWidget()
        hbox_set = QHBoxLayout(button_set)
        hbox_set.setContentsMargins(0, 0, 0, 0)
        self.btn_cancel = btn_cancel = QPushButton("Cancel")
        self.btn_add = btn_add = QPushButton("Add")
        btn_add.setDisabled(True)
        btn_cancel.setFixedSize(96, 24)
        btn_add.setFixedSize(96, 24)
        hbox_set.addStretch()
        hbox_set.addWidget(btn_cancel)
        hbox_set.addWidget(btn_add)

        vbox.addWidget(action_bar)
        vbox.addWidget(group_info)
        vbox.addStretch()
        vbox.addWidget(button_set)

    def config_logic(self):
        self.edit_name.textChanged.connect(self.var_changed)
        self.btn_cancel.clicked.connect(self.close)
        self.btn_check.clicked.connect(self.check_var)
        self.btn_add.clicked.connect(self.add_var)

    def var_changed(self, text):
        self.btn_add.setDisabled(True)

        if text:
            self.btn_check.setDisabled(False)
        else:
            self.btn_check.setDisabled(True)

    def check_var(self):
        name = self.edit_name.text()
        try:
            value = self.env._get_variables([name])[name]
            min, max = self.env._get_bounds([name])[name]

            self.edit_value.edit.setText(str(value))
            self.edit_min.edit.setText(str(min))
            self.edit_max.edit.setText(str(max))

            self.btn_add.setDisabled(False)
        except Exception:
            self.edit_value.edit.setText("")
            self.edit_min.edit.setText("")
            self.edit_max.edit.setText("")
            QMessageBox.critical(self, "Error!", f"Variable {name} cannot be found!")

            self.btn_add.setDisabled(True)

    def add_var(self):
        name = self.edit_name.text()
        min = float(self.edit_min.edit.text())
        max = float(self.edit_max.edit.text())

        code = self.callback(name, min, max)
        if code == 0:
            self.close()



import argparse

from badger.log import config_log

config_log()
from badger.actions import show_info  # noqa: E402
from badger.actions.doctor import self_check  # noqa: E402
from badger.actions.routine import show_routine  # noqa: E402
from badger.actions.generator import show_generator  # noqa: E402
from badger.actions.env import show_env  # noqa: E402
from badger.actions.install import plugin_install  # noqa: E402
from badger.actions.uninstall import plugin_remove  # noqa: E402
from badger.actions.intf import show_intf  # noqa: E402
from badger.actions.run import run_routine  # noqa: E402
from badger.actions.config import config_settings  # noqa: E402


def main():
    # Create the top-level parser
    parser = argparse.ArgumentParser(description="Badger the optimizer")
    parser.add_argument("-g", "--gui", action="store_true", help="launch the GUI")
    parser.add_argument(
        "-ga", "--gui-acr", action="store_true", help="launch the GUI for ACR"
    )
    parser.add_argument(
        "-l",
        "--log",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
        default="WARNING",
        const="WARNING",
        nargs="?",
        help="change the log level",
    )
    parser.add_argument(
        "-cf",
        "--config_filepath",
        type=str,
        default=None,
        help="Path to the config file",
    )
    parser.set_defaults(func=show_info)
    subparsers = parser.add_subparsers(help="Badger commands help")

    # Parser for the 'doctor' command
    parser_doctor = subparsers.add_parser("doctor", help="Badger status self-check")
    parser_doctor.add_argument(
        "-r", "--reset", action="store_true", help="reset Badger to factory settings"
    )
    parser_doctor.set_defaults(func=self_check)

    # Parser for the 'routine' command
    parser_routine = subparsers.add_parser("routine", help="Badger routines")
    parser_routine.add_argument("routine_name", nargs="?", type=str, default=None)
    parser_routine.add_argument(
        "-r", "--run", action="store_true", help="run the routine"
    )
    parser_routine.add_argument(
        "-y", "--yes", action="store_true", help="run the routine without confirmation"
    )
    parser_routine.add_argument(
        "-v",
        "--verbose",
        type=int,
        choices=[0, 1, 2],
        default=2,
        const=2,
        nargs="?",
        help="verbose level of optimization progress",
    )
    parser_routine.set_defaults(func=show_routine)

    # Parser for the 'generator' command
    parser_generator = subparsers.add_parser("generator", help="Badger generators")
    parser_generator.add_argument("generator_name", nargs="?", type=str, default=None)
    parser_generator.set_defaults(func=show_generator)

    # Parser for the 'intf' command
    parser_intf = subparsers.add_parser("intf", help="Badger interfaces")
    parser_intf.add_argument("intf_name", nargs="?", type=str, default=None)
    parser_intf.set_defaults(func=show_intf)

    # Parser for the 'env' command
    parser_env = subparsers.add_parser("env", help="Badger environments")
    parser_env.add_argument("env_name", nargs="?", type=str, default=None)
    parser_env.set_defaults(func=show_env)

    # Parser for the 'install' command
    parser_inst = subparsers.add_parser("install", help="Badger install plugin")
    parser_inst.add_argument("plugin_type", nargs="?", type=str, default=None)
    parser_inst.add_argument("plugin_specific", nargs="?", type=str, default=None)
    parser_inst.set_defaults(func=plugin_install)

    # Parser for the 'remove' command
    parser_remove = subparsers.add_parser("remove", help="Badger remove plugin")
    parser_remove.add_argument("plugin_type", nargs="?", type=str, default=None)
    parser_remove.add_argument("plugin_specific", nargs="?", type=str, default=None)
    parser_remove.set_defaults(func=plugin_remove)

    # Parser for the 'run' command
    parser_run = subparsers.add_parser("run", help="run routines")
    parser_run.add_argument("-a", "--generator", required=True, help="generator to use")
    parser_run.add_argument(
        "-ap", "--generator_params", help="parameters for the generator"
    )
    parser_run.add_argument("-e", "--env", required=True, help="environment to use")
    parser_run.add_argument(
        "-ep", "--env_params", help="parameters for the environment"
    )
    parser_run.add_argument(
        "-c", "--config", required=True, help="config for the routine"
    )
    parser_run.add_argument(
        "-s", "--save", nargs="?", const="", help="the routine name to be saved"
    )
    parser_run.add_argument(
        "-y", "--yes", action="store_true", help="run the routine without confirmation"
    )
    parser_run.add_argument(
        "-v",
        "--verbose",
        type=int,
        choices=[0, 1, 2],
        default=2,
        const=2,
        nargs="?",
        help="verbose level of optimization progress",
    )
    parser_run.set_defaults(func=run_routine)

    # Parser for the 'config' command
    parser_config = subparsers.add_parser("config", help="Badger configurations")
    parser_config.add_argument("key", nargs="?", type=str, default=None)
    parser_config.set_defaults(func=config_settings)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import sys

import pytest


if __name__ == "__main__":
    # Show output results from every test function
    # Show the message output for skipped and expected failures
    args = ["-v", "-vrxs"]

    # Add extra arguments
    if len(sys.argv) > 1:
        args.extend(sys.argv[1:])

    # Show coverage
    if "--show-cov" in args:
        args.extend(["--cov=badger", "--cov-report", "term-missing"])
        args.remove("--show-cov")

    print("pytest arguments: {}".format(args))
    print(f"Running tests on Python {sys.version}")
    sys.exit(pytest.main(args))



import logging
import time
import traceback

import pandas as pd
import torch  # noqa: F401. For converting dtype str to torch object.
from PyQt5.QtCore import pyqtSignal, QObject, QTimer

from badger.errors import BadgerRunTerminated
from badger.tests.utils import get_current_vars
from badger.routine import calculate_variable_bounds, calculate_initial_points
from badger.settings import init_settings
from badger.gui.default.components.process_manager import ProcessManager
from badger.routine import Routine
from badger.errors import BadgerError

logger = logging.getLogger(__name__)


class BadgerRoutineSignals(QObject):
    env_ready = pyqtSignal(list)
    finished = pyqtSignal()
    progress = pyqtSignal(object)
    error = pyqtSignal(Exception)
    info = pyqtSignal(str)
    states = pyqtSignal(str)


class BadgerRoutineSubprocess:
    """
    This class takes the users chosen routine and then grabs a subprocess to run the routine
    using code in core_subprocess.py.
    """

    def __init__(
        self,
        process_manager: ProcessManager,
        routine: Routine = None,
        routine_filename: str = None,
        save: bool = False,
        verbose: int = 2,
        use_full_ts: bool = False,
        testing: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        routine: Routine
            Defined routine for runner
        save: bool
            Flag to enable saving to database
        verbose: int, default: 2
            Verbostiy level (higher is more output)
        use_full_ts: bool
            If true use full time stamp info when dumping to database
        """
        super().__init__()
        # Signals should belong to instance rather than class
        # Since there could be multiple runners running in parallel
        self.signals = BadgerRoutineSignals()
        self.process_manager = process_manager
        self.routine = routine
        self.routine_filename = routine_filename
        self.run_filename = None
        self.states = None  # system states to be saved at start of a run
        self.save = save
        self.verbose = verbose
        self.use_full_ts = use_full_ts
        self.termination_condition = (
            None  # additional option to control the optimization flow
        )
        self.start_time = None  # track the time cost of the run
        self.last_dump_time = None  # track the time the run data got dumped
        self.data_and_error_queue = None
        self.stop_event = None
        self.pause_event = None
        self.routine_process = None
        self.is_killed = False
        self.interval = 100
        self.testing = testing
        self.config_singleton = init_settings()

    def set_termination_condition(self, termination_condition: dict) -> None:
        """
        Setter method for the termination condition.

        Parameters
        ----------
        termination_condition : dict
        """
        self.termination_condition = termination_condition

    def run(self) -> None:
        """
        This method starts up the routine.
        The method grabs a subprocess from self.process_manager queue.
        Then the method unpauses the process along with passing the selected routine name.
        """

        self.start_time = time.time()
        self.last_dump_time = None  # reset the timer

        # Patch for converting dtype str to torch object
        try:
            dtype = self.routine.generator.turbo_controller.tkwargs["dtype"]
            self.routine.generator.turbo_controller.tkwargs["dtype"] = eval(dtype)
        except AttributeError:
            pass
        except KeyError:
            pass
        except TypeError:
            pass

        self.routine.data = None  # reset data
        # Recalculate the bounds and initial points if asked
        if (
            self.config_singleton.read_value("AUTO_REFRESH")
            and self.routine.relative_to_current
        ):
            variables_updated = calculate_variable_bounds(
                self.routine.vrange_limit_options,
                self.routine.vocs,
                self.routine.environment,
            )

            self.routine.vocs.variables = variables_updated

            init_points = calculate_initial_points(
                self.routine.initial_point_actions,
                self.routine.vocs,
                self.routine.environment,
            )
            try:
                init_points = pd.DataFrame(init_points)
            except IndexError:
                init_points = pd.DataFrame(init_points, index=[0])

            self.routine.initial_points = init_points

        try:
            self.save_init_vars()
            process_with_args = self.process_manager.remove_from_queue()
            self.routine_process = process_with_args["process"]
            self.stop_event = process_with_args["stop_event"]
            self.pause_event = process_with_args["pause_event"]
            self.data_and_error_queue = process_with_args["data_queue"]
            self.evaluate_queue = process_with_args["evaluate_queue"]
            self.wait_event = process_with_args["wait_event"]

            arg_dict = {
                "routine_id": self.routine.id,
                "routine_filename": self.routine_filename,
                "routine_name": self.routine.name,
                "variable_ranges": self.routine.vocs.variables,
                "initial_points": self.routine.initial_points,
                "evaluate": True,
                "archive": self.save,
                "termination_condition": self.termination_condition,
                "start_time": self.start_time,
                "testing": self.testing,
            }

            self.data_and_error_queue.put(arg_dict)
            self.wait_event.set()
            self.pause_event.set()
            self.setup_timer()
            # self.signals.finished.emit(self.routine.states)

            self.routine.data = None  # reset data
            # Recalculate the bounds and initial points if asked
            if (
                self.config_singleton.read_value("AUTO_REFRESH")
                and self.routine.relative_to_current
            ):
                variables_updated = calculate_variable_bounds(
                    self.routine.vrange_limit_options,
                    self.routine.vocs,
                    self.routine.environment,
                )

                self.routine.vocs.variables = variables_updated

                init_points = calculate_initial_points(
                    self.routine.initial_point_actions,
                    self.routine.vocs,
                    self.routine.environment,
                )
                try:
                    init_points = pd.DataFrame(init_points)
                except IndexError:
                    init_points = pd.DataFrame(init_points, index=[0])

                self.routine.initial_points = init_points

        except BadgerRunTerminated as e:
            self.signals.finished.emit()
            self.signals.info.emit(str(e))
        except Exception as e:
            traceback_info = traceback.format_exc()
            e._details = traceback_info
            self.signals.finished.emit()
            self.signals.error.emit(e)

    def setup_timer(self) -> None:
        """
        This sets up a QTimer to check for updates from the subprocess.
        The clock checks are every 100 miliseconds.
        """
        self.timer = QTimer()
        self.timer.timeout.connect(self.check_queue)
        self.timer.setInterval(self.interval)
        self.timer.start()

    def check_queue(self) -> None:
        """
        This method checks the subprocess for updates in the evaluate_queue.
        It also checks the self.data_and_error_queue to see if an exception was thrown during the routine.
        It is called by a QTimer every 100 miliseconds.
        """
        if self.evaluate_queue[1].poll():
            while self.evaluate_queue[1].poll():
                results = self.evaluate_queue[1].recv()
                self.after_evaluate(results[0])
                self.routine.generator = results[1]

        if not self.data_and_error_queue.empty():
            try:
                error_title, error_traceback = self.data_and_error_queue.get()
                BadgerError(error_title, error_traceback)
            except ValueError:  # seems to only occur in tests
                pass

        if not self.routine_process.is_alive():
            self.close()
            self.evaluate_queue[1].close()

    def after_evaluate(self, results: pd.DataFrame) -> None:
        """
        This method emits updates sent over from the subprocess running the routine on the evaluate_queue.

        Parameters
        ----------
        results : DataFrame
        """
        if self.timer.isActive():
            self.signals.progress.emit(results)

    def save_init_vars(self) -> None:
        """
        Emits the intital variables in the env_ready signal.
        """
        init_vars = get_current_vars(self.routine)
        self.signals.env_ready.emit(init_vars)

    def stop_routine(self) -> None:
        """
        This method will attempt to close the subprocess running the routine.
        If the process does not close withing the timeout time (0.1 seconds)
        then the method will terminate the process.
        The method then emits a signal that the process has been stopped.
        """
        self.stop_event.set()
        self.routine_process.join(timeout=0.1)

        if self.routine_process.is_alive():
            self.routine_process.terminate()
            self.routine_process.join()

        self.close()

    def ctrl_routine(self, pause: bool) -> None:
        """
        This method will pause and unpause the routine.
        This is accomplished by a multiprocessing Event which when set will pause the subprocess.

        Parameters
        ----------
        pause : bool
        """
        if pause:
            self.pause_event.clear()
        else:
            self.pause_event.set()

    def close(self) -> None:
        self.timer.stop()
        self.signals.finished.emit()



import numpy as np


def none(data):
    return data


def median(data):
    return np.median(data)


def std_deviation(data):
    return np.std(data)


def median_deviation(data):
    median = np.median(data)

    return np.median(np.abs(data - median))


def max(data):
    return np.max(data)


def min(data):
    return np.min(data)


def percent_80(data):
    return np.percentile(data, 80)


def percent_20(data):
    return np.percentile(data, 20)


def avg_mean(data):
    percentile = np.percentile(data, 50)

    return np.mean(data[data > percentile])


def mean(data):
    return np.mean(data)



import pytest


def test_find_env():
    from badger.factory import list_env, get_env

    assert len(list_env()) == 2

    _, configs = get_env("test")
    assert configs["name"] == "test"
    assert configs["version"] == "1.0"


def test_get_params():
    from badger.factory import get_env

    _, configs = get_env("test")
    params = configs["params"]

    assert params == {"delay": 0.0, "flag": 0}


def test_set_params():
    from badger.factory import get_env, get_intf

    Interface, _ = get_intf("test")
    intf = Interface()

    # Test pass params only
    Environment, _ = get_env("test")
    env = Environment(flag=1)

    assert env.interface is None
    assert env.flag == 1

    # Test pass interface
    env = Environment(interface=intf)
    assert env.interface
    assert env.flag == 0


def test_list_variables():
    from badger.factory import get_env

    Environment, _ = get_env("test")
    env = Environment()

    variables = {f"x{i}": [-1, 1] for i in range(20)}
    assert env.variables == variables

    assert env.variable_names == list(variables.keys())


def test_list_observables():
    from badger.factory import get_env

    Environment, _ = get_env("test")
    env = Environment()

    assert env.observables == ["f", "c"]


def test_get_variables():
    from badger.factory import get_env, get_intf
    from badger.errors import (
        BadgerNoInterfaceError,
    )

    Interface, _ = get_intf("test")
    intf = Interface()

    # Without interface
    Environment, _ = get_env("test")
    env = Environment()

    with pytest.raises(BadgerNoInterfaceError):
        env._get_variables(["x1", "x2"])

    # With interface
    env.interface = intf

    variable_outputs = env._get_variables(["x1", "x2"])
    assert variable_outputs == {"x1": 0.5, "x2": 0.5}

    # Test getting variables not defined in env
    # with pytest.raises(Exception) as e:
    #     variable_outputs = env._get_variables(["x19", "x20"])

    # assert e.type == BadgerInterfaceChannelError
    # assert "x19" not in str(e.value)
    # assert "x20" in str(e.value)
    # assert "Getting them through interface is not allowed" in str(e.value)


def test_set_variables():
    from badger.factory import get_env, get_intf
    from badger.errors import (
        BadgerEnvVarError,
        BadgerNoInterfaceError,
    )

    Interface, _ = get_intf("test")
    intf = Interface()

    # Without interface
    Environment, _ = get_env("test")
    env = Environment()

    variable_inputs = {"x1": 1, "x2": -1}
    with pytest.raises(BadgerNoInterfaceError):
        env._set_variables(variable_inputs)

    # With interface
    env.interface = intf

    env._set_variables(variable_inputs)
    variable_outputs = env._get_variables(["x1", "x2"])
    assert variable_outputs == {"x1": 1, "x2": -1}

    # Test setting variables not defined in env
    # variable_inputs_undef = {"x21": 1, "x22": -1}
    # with pytest.raises(Exception) as e:
    #     env._set_variables(variable_inputs_undef)

    # assert e.type == BadgerInterfaceChannelError
    # assert "x21" in str(e.value)
    # assert "Setting them through interface is not allowed" in str(e.value)

    # Test setting variables out of range
    variable_inputs_out_range = {"x1": 0, "x2": -2}
    with pytest.raises(Exception) as e:
        env._set_variables(variable_inputs_out_range)

    assert e.type == BadgerEnvVarError  # getting crazy w/ custom error
    assert "x1" not in str(e.value)
    assert "x2" in str(e.value)
    assert "outside its bounds" in str(e.value)

    variable_outputs = env._get_variables(["x1", "x2"])
    assert variable_outputs == {"x1": 1, "x2": -1}  # values shouldn't change


def test_get_observables():
    from badger.factory import get_env, get_intf
    from badger.errors import BadgerNoInterfaceError

    Interface, _ = get_intf("test")
    intf = Interface()

    # Without interface
    Environment, _ = get_env("test")
    env = Environment()

    with pytest.raises(BadgerNoInterfaceError):
        env._get_observables(["f"])

    # With interface
    env.interface = intf

    variable_outputs = env._get_observables(["f"])
    assert variable_outputs == {"f": 0.5}

    variable_inputs = {"x1": 1, "x2": -1}
    env._set_variables(variable_inputs)
    variable_outputs = env._get_observables(["f"])
    assert variable_outputs == {"f": 6.5}



from PyQt5 import QtCore, QtGui


def format(color, style=""):
    """Return a QTextCharFormat with the given attributes."""
    _color = QtGui.QColor()
    _color.setNamedColor(color)

    _format = QtGui.QTextCharFormat()
    _format.setForeground(_color)
    if "bold" in style:
        _format.setFontWeight(QtGui.QFont.Bold)
    if "italic" in style:
        _format.setFontItalic(True)

    return _format


# Syntax styles that can be shared by all languages
STYLES = {
    "keyword": format("#81A1C1"),
    "operator": format("#81A1C1"),
    "brace": format("white"),
    "defclass": format("#8FBCBB", "bold"),
    "string": format("#A3BE8C"),
    "string2": format("#A3BE8C"),
    "comment": format("#4C566A", "italic"),
    "self": format("white", "italic"),
    "numbers": format("#B48EAD"),
}


class PythonHighlighter(QtGui.QSyntaxHighlighter):
    """Syntax highlighter for the Python language."""

    # Python keywords
    keywords = [
        "and",
        "assert",
        "break",
        "class",
        "continue",
        "def",
        "del",
        "elif",
        "else",
        "except",
        "exec",
        "finally",
        "for",
        "from",
        "global",
        "if",
        "import",
        "in",
        "is",
        "lambda",
        "not",
        "or",
        "pass",
        "print",
        "raise",
        "return",
        "try",
        "while",
        "yield",
        "None",
        "True",
        "False",
        "as",
    ]

    # Python operators
    operators = [
        "=",
        # Comparison
        "==",
        "!=",
        "<",
        "<=",
        ">",
        ">=",
        # Arithmetic
        "\+",
        "-",
        "\*",
        "/",
        "//",
        "\%",
        "\*\*",
        # In-place
        "\+=",
        "-=",
        "\*=",
        "/=",
        "\%=",
        # Bitwise
        "\^",
        "\|",
        "\&",
        "\~",
        ">>",
        "<<",
    ]

    # Python braces
    braces = [
        "\{",
        "\}",
        "\(",
        "\)",
        "\[",
        "\]",
    ]

    def __init__(self, parent: QtGui.QTextDocument) -> None:
        super().__init__(parent)

        # Multi-line strings (expression, flag, style)
        self.tri_single = (QtCore.QRegExp("'''"), 1, STYLES["string2"])
        self.tri_double = (QtCore.QRegExp('"""'), 2, STYLES["string2"])

        rules = []

        # Keyword, operator, and brace rules
        rules += [
            (r"\b%s\b" % w, 0, STYLES["keyword"]) for w in PythonHighlighter.keywords
        ]
        rules += [
            (r"%s" % o, 0, STYLES["operator"]) for o in PythonHighlighter.operators
        ]
        rules += [(r"%s" % b, 0, STYLES["brace"]) for b in PythonHighlighter.braces]

        # All other rules
        rules += [
            # 'self'
            (r"\bself\b", 0, STYLES["self"]),
            # 'def' followed by an identifier
            (r"\bdef\b\s*(\w+)", 1, STYLES["defclass"]),
            # 'class' followed by an identifier
            (r"\bclass\b\s*(\w+)", 1, STYLES["defclass"]),
            # Numeric literals
            (r"\b[+-]?[0-9]+[lL]?\b", 0, STYLES["numbers"]),
            (r"\b[+-]?0[xX][0-9A-Fa-f]+[lL]?\b", 0, STYLES["numbers"]),
            (r"\b[+-]?[0-9]+(?:\.[0-9]+)?(?:[eE][+-]?[0-9]+)?\b", 0, STYLES["numbers"]),
            # Double-quoted string, possibly containing escape sequences
            (r'"[^"\\]*(\\.[^"\\]*)*"', 0, STYLES["string"]),
            # Single-quoted string, possibly containing escape sequences
            (r"'[^'\\]*(\\.[^'\\]*)*'", 0, STYLES["string"]),
            # From '#' until a newline
            (r"#[^\n]*", 0, STYLES["comment"]),
        ]

        # Build a QRegExp for each pattern
        self.rules = [(QtCore.QRegExp(pat), index, fmt) for (pat, index, fmt) in rules]

    def highlightBlock(self, text):
        """Apply syntax highlighting to the given block of text."""
        self.tripleQuoutesWithinStrings = []
        # Do other syntax formatting
        for expression, nth, format in self.rules:
            index = expression.indexIn(text, 0)
            if index >= 0:
                # if there is a string we check
                # if there are some triple quotes within the string
                # they will be ignored if they are matched again
                if expression.pattern() in [
                    r'"[^"\\]*(\\.[^"\\]*)*"',
                    r"'[^'\\]*(\\.[^'\\]*)*'",
                ]:
                    innerIndex = self.tri_single[0].indexIn(text, index + 1)
                    if innerIndex == -1:
                        innerIndex = self.tri_double[0].indexIn(text, index + 1)

                    if innerIndex != -1:
                        tripleQuoteIndexes = range(innerIndex, innerIndex + 3)
                        self.tripleQuoutesWithinStrings.extend(tripleQuoteIndexes)

            while index >= 0:
                # skipping triple quotes within strings
                if index in self.tripleQuoutesWithinStrings:
                    index += 1
                    expression.indexIn(text, index)
                    continue

                # We actually want the index of the nth match
                index = expression.pos(nth)
                length = len(expression.cap(nth))
                self.setFormat(index, length, format)
                index = expression.indexIn(text, index + length)

        self.setCurrentBlockState(0)

        # Do multi-line strings
        in_multiline = self.match_multiline(text, *self.tri_single)
        if not in_multiline:
            in_multiline = self.match_multiline(text, *self.tri_double)

    def match_multiline(self, text, delimiter, in_state, style):
        """Do highlighting of multi-line strings. ``delimiter`` should be a
        ``QRegExp`` for triple-single-quotes or triple-double-quotes, and
        ``in_state`` should be a unique integer to represent the corresponding
        state changes when inside those strings. Returns True if we're still
        inside a multi-line string when this function is finished.
        """
        # If inside triple-single quotes, start at 0
        if self.previousBlockState() == in_state:
            start = 0
            add = 0
        # Otherwise, look for the delimiter on this line
        else:
            start = delimiter.indexIn(text)
            # skipping triple quotes within strings
            if start in self.tripleQuoutesWithinStrings:
                return False
            # Move past this match
            add = delimiter.matchedLength()

        # As long as there's a delimiter match on this line...
        while start >= 0:
            # Look for the ending delimiter
            end = delimiter.indexIn(text, start + add)
            # Ending delimiter on this line?
            if end >= add:
                length = end - start + add + delimiter.matchedLength()
                self.setCurrentBlockState(0)
            # No; multi-line string
            else:
                self.setCurrentBlockState(in_state)
                length = len(text) - start + add
            # Apply formatting
            self.setFormat(start, length, style)
            # Look for the next match
            start = delimiter.indexIn(text, start + length)

        # Return True if still inside a multi-line string, False otherwise
        if self.currentBlockState() == in_state:
            return True
        else:
            return False



def test_data_viewer_simple(qtbot):
    from badger.gui.default.components.analysis_extensions import AnalysisExtension
    from badger.tests.utils import create_routine

    routine = create_routine()
    routine.evaluate_data(routine.initial_points)

    data_viewer = AnalysisExtension()
    data_viewer.update_window(routine)



import pickle
from abc import ABC, abstractmethod
from typing import Any, ClassVar, Dict, List

from pydantic import BaseModel

from badger.utils import curr_ts


def log(func):
    def func_log(*args, **kwargs):
        if func.__name__ == "set_values":
            if "channel_inputs" in kwargs.keys():
                channel_inputs = kwargs["channel_inputs"]
            else:
                channel_inputs = args[1]
            args[0]._logs.append(
                {
                    "timestamp": curr_ts().timestamp(),
                    "action": "set_values",
                    "channel_inputs": channel_inputs,
                }
            )

            return func(*args, **kwargs)
        elif func.__name__ == "get_values":
            channel_outputs = func(*args, **kwargs)
            args[0]._logs.append(
                {
                    "timestamp": curr_ts().timestamp(),
                    "action": "get_values",
                    "channel_outputs": channel_outputs,
                }
            )

            return channel_outputs
        else:
            return func(*args, **kwargs)

    return func_log


class Interface(BaseModel, ABC):
    name: ClassVar[str]
    # Put interface params here
    # params: float = Field(..., description='Example intf parameter')

    # Private variables
    _logs: List[Dict] = []  # TODO: Add a property for it?

    def start_recording(self):
        self._logs = []

    def stop_recording(self, filename):
        # Do not create the pickle file if log is empty
        # This usually happens when the log decorator is not used in the
        # specific interface
        if not self._logs:
            return

        with open(filename, "wb") as f:
            pickle.dump(self._logs, f)

        self._logs = []

    def dump_recording(self, filename):
        # Dump the logs to disk w/o cleaning up the log history
        if not self._logs:
            return

        with open(filename, "wb") as f:
            pickle.dump(self._logs, f)

    # Environment should only call this method to get channels
    @abstractmethod
    def get_values(self, channel_names: List[str]) -> Dict[str, Any]:
        pass

    # Environment should only call this method to set channels
    @abstractmethod
    def set_values(self, channel_inputs: Dict[str, Any]):
        pass

    def reset_interface(self):
        """
        Called after the application forks (i.e. after spawning a new multiprocess.Process)
        Subclasses should use this to reset any undesirable process wide state
        """
        pass

    def get_value(self, channel_name: str, **kwargs) -> Any:
        return self.get_values([channel_name], **kwargs)[channel_name]

    def set_value(self, channel_name: str, channel_value, **kwargs):
        return self.set_values({channel_name: channel_value}, **kwargs)
